# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Provides class for horizontal two-dimensional spatio-temporal grid.
"""

from typing import Tuple
from collections.abc import Hashable

import numpy as np
# Import package to calculate points inside a polygon
try:
    from matplotlib import path
except ImportError:
    print("An error occurred when importing matplotlib.path! "
          "Some functionality in GeoGrid class might not be available.")

from ._ext.types import to_cy, FIELD
from ._ext.numerics import _calculate_angular_distance
from .cache import Cached
from .grid import Grid


class GeoGrid(Grid):
    """
    Encapsulates a horizontal two-dimensional spatio-temporal grid on the
    sphere.

    The spatial grid points can be arbitrarily distributed, which is useful
    for representing station data or geodesic grids.
    """

    #
    #  Definitions of internal methods
    #

    def __init__(self, time_seq: np.ndarray,
                 lat_seq: np.ndarray, lon_seq: np.ndarray,
                 silence_level: int = 0):
        """
        Initialize an instance of GeoGrid.

        :type time_seq: 1D Numpy array [time]
        :arg time_seq: The increasing sequence of temporal sampling points.

        :type lat_seq: 1D Numpy array [index]
        :arg lat_seq: The sequence of latitudinal sampling points.

        :type lon_seq: 1D Numpy array [index]
        :arg lon_seq: The sequence of longitudinal sampling points.

        :type silence_level: number (int)
        :arg silence_level: The inverse level of verbosity of the object.
        """
        Grid.__init__(self, time_seq, np.vstack((lat_seq, lon_seq)),
                      silence_level)

    def __cache_state__(self) -> Tuple[Hashable, ...]:
        return Grid.__cache_state__(self)

    def __str__(self):
        """
        Return a string representation of the GeoGrid object.
        """
        return (f"GeoGrid: {self._grid_size['space']} grid points, "
                f"{self._grid_size['time']} timesteps.")

    #
    #  Functions for loading and saving the Grid object
    #

    def save_txt(self, filename):
        """
        Save the GeoGrid object to text files.

        The latitude, longitude and time sequences are stored in three separate
        text files.

        :arg str filename: The name of the files where Grid object is stored
            (excluding ending).
        """
        #  Gather sequences
        lat_seq = self.lat_sequence()
        lon_seq = self.lon_sequence()
        time_seq = self.grid()["time"]

        #  Store as text files
        try:
            np.savetxt(filename + "_lat.txt", lat_seq)
            np.savetxt(filename + "_lon.txt", lon_seq)
            np.savetxt(filename + "_time.txt", time_seq)
        except IOError:
            print("An error occurred while saving Grid instance to "
                  f"text files {filename}")

    @staticmethod
    def LoadTXT(filename):
        """
        Return a GeoGrid object stored in text files.

        The latitude, longitude and time sequences are loaded from three
        separate text files.

        :arg str filename: The name of the files where the GeoGrid object is
            stored (excluding endings).
        :rtype: Grid object
        :return: :class:`GeoGrid` instance.
        """
        try:
            lat_seq = np.loadtxt(filename + "_lat.txt")
            lon_seq = np.loadtxt(filename + "_lon.txt")
            time_seq = np.loadtxt(filename + "_time.txt")
        except IOError:
            print("An error occurred while loading Grid instance from "
                  f"text files {filename}")

        return GeoGrid(time_seq, lat_seq, lon_seq)

    #
    #  Alternative constructors and Grid generation methods
    #

    @staticmethod
    def SmallTestGrid():
        """
        Return test grid of 6 spatial grid points with 10 temporal sampling
        points each.

        :rtype: GeoGrid instance
        :return: a GeoGrid instance for testing purposes.
        """
        return GeoGrid(time_seq=np.arange(10),
                       lat_seq=np.array([0, 5, 10, 15, 20, 25]),
                       lon_seq=np.array([2.5, 5., 7.5, 10., 12.5, 15.]),
                       silence_level=2)

    @staticmethod
    def RegularGrid(time_seq, space_grid, silence_level=0):
        """
        Initialize an instance of a regular grid.

        **Examples:**

        >>> GeoGrid.RegularGrid(
        ...      time_seq=np.arange(2),
        ...      space_grid=(np.array([0.,5.]),
        ...                  np.array([1.,2.])),
        ...      silence_level=2).lat_sequence()
        array([ 0.,  0.,  5.,  5.], dtype=float32)
        >>> GeoGrid.RegularGrid(
        ...      time_seq=np.arange(2),
        ...      space_grid=(np.array([0.,5.]),
        ...                  np.array([1.,2.])),
        ...      silence_level=2).lon_sequence()
        array([ 1.,  2.,  1.,  2.], dtype=float32)

        :type time_seq: 1D Numpy array [time]
        :arg time_seq: The increasing sequence of temporal sampling points.
        :type space_grid: tuple or list of two 1D Numpy arrays
            ([n_lat], [n_lon])
        :arg space_grid: The spatial grid, consisting of the latitudinal
            and the longitudinal grid.
        :type silence_level: number (int)
        :arg silence_level: The inverse level of verbosity of the object.

        :rtype: GeoGrid object
        :return: :class:`GeoGrid` instance.
        """
        try:
            (lat_grid, lon_grid) = space_grid
        except ValueError as e:
            raise ValueError("'space_grid' must be a tuple or list of two "
                             "items: lat_grid, lon_grid") from e

        #  Generate sequence of latitudes and longitudes for all nodes
        lat_seq, lon_seq = GeoGrid.coord_sequence_from_rect_grid(lat_grid,
                                                                 lon_grid)

        #  Return instance of Grid
        return GeoGrid(time_seq, lat_seq, lon_seq, silence_level)

    #
    #  Definitions of grid related functions
    #

    @staticmethod
    def coord_sequence_from_rect_grid(lat_grid, lon_grid):
        """
        Return the sequences of latitude and longitude for a regular and
        rectangular grid.

        **Example:**

        >>> GeoGrid.coord_sequence_from_rect_grid(
        ...     lat_grid=np.array([0.,5.]), lon_grid=np.array([1.,2.]))
        (array([ 0.,  0.,  5.,  5.]), array([ 1.,  2.,  1.,  2.]))

        :type lat_grid: 1D Numpy array [lat]
        :arg lat_grid: The grid's latitudinal sampling points.

        :type lon_grid: 1D Numpy array [lon]
        :arg lon_grid: The grid's longitudinal sampling points.

        :rtype: tuple of two 1D Numpy arrays [index]
        :return: the coordinates of all nodes in the grid.
        """
        space_seq = Grid.coord_sequence_from_rect_grid([lat_grid, lon_grid])

        #  Return results as a tuple
        return (space_seq[0], space_seq[1])

    def lat_sequence(self):
        """
        Return the sequence of latitudes for all nodes.

        **Example:**

        >>> GeoGrid.SmallTestGrid().lat_sequence()
        array([  0.,   5.,  10.,  15.,  20.,  25.], dtype=float32)

        :rtype: 1D Numpy array [index]
        :return: the sequence of latitudes for all nodes.
        """
        return self.sequence(0)

    def lon_sequence(self):
        """
        Return the sequence of longitudes for all nodes.

        **Example:**

        >>> GeoGrid.SmallTestGrid().lon_sequence()
        array([  2.5,   5. ,   7.5,  10. ,  12.5,  15. ], dtype=float32)

        :rtype: 1D Numpy array [index]
        :return: the sequence of longitudes for all nodes.
        """
        return self.sequence(1)

    def convert_lon_coordinates(self, lon_seq):
        """
        Return longitude coordinates in the system
        -180 deg W <= lon <= +180 deg O for all nodes.

        Accepts longitude coordinates in the system 0 deg <= lon <= 360 deg.
        0 deg corresponds to Greenwich, England.

        **Example:**

        >>> GeoGrid.SmallTestGrid().convert_lon_coordinates(
        ...     np.array([10.,350.,20.,340.,170.,190.]))
        array([  10.,  -10.,   20.,  -20.,  170., -170.])

        :type lon_seq: 1D Numpy array [index]
        :arg lon_seq: Sequence of longitude coordinates.

        :rtype: 1D Numpy array [index]
        :return: the converted longitude coordinates for all nodes.
        """
        new_lon_grid = np.empty(self.N)

        for i in range(self.N):
            if lon_seq[i] > 180.:
                new_lon_grid[i] = lon_seq[i] - 360.
            else:
                new_lon_grid[i] = lon_seq[i]

        return new_lon_grid

    def node_number(self, lat_node, lon_node):
        """
        Return the index of the closest node given geographical coordinates.

        **Example:**

        >>> GeoGrid.SmallTestGrid().node_number(lat_node=14., lon_node=9.)
        3

        :type lat_node: number (float)
        :arg lat_node: The latitude coordinate.

        :type lon_node: number (float)
        :arg lon_node: The longitude coordinate.

        :rtype: number (int)
        :return: the closest node's index.
        """
        # Get sequences of cosLat, sinLat, cosLon and sinLon for all nodes
        cos_lat = self.cos_lat()
        sin_lat = self.sin_lat()
        cos_lon = self.cos_lon()
        sin_lon = self.sin_lon()

        sin_lat_v = np.sin(lat_node * np.pi / 180)
        cos_lat_v = np.cos(lat_node * np.pi / 180)
        sin_lon_v = np.sin(lon_node * np.pi / 180)
        cos_lon_v = np.cos(lon_node * np.pi / 180)

        #  Calculate angular distance from the given coordinate to all
        #  other nodes
        expr = sin_lat*sin_lat_v + cos_lat*cos_lat_v * (sin_lon*sin_lon_v
                                                        + cos_lon*cos_lon_v)

        #  Correct for rounding errors
        expr[expr < -1.] = -1.
        expr[expr > 1.] = 1.

        angdist = np.arccos(expr)

        #  Get index of closest node
        n_node = angdist.argmin()

        return n_node

    def cos_lat(self):
        """
        Return the sequence of cosines of latitude for all nodes.

        **Example:**

        >>> r(GeoGrid.SmallTestGrid().cos_lat()[:2])
        array([ 1. , 0.9962])

        :rtype: 1D Numpy array [index]
        :return: the cosine of latitudes for all nodes.
        """
        return np.cos(self.lat_sequence() * np.pi / 180)

    def sin_lat(self):
        """
        Return the sequence of sines of latitude for all nodes.

        **Example:**

        >>> r(GeoGrid.SmallTestGrid().sin_lat()[:2])
        array([ 0. , 0.0872])

        :rtype: 1D Numpy array [index]
        :return: the sine of latitudes for all nodes.
        """
        return np.sin(self.lat_sequence() * np.pi / 180)

    def cos_lon(self):
        """
        Return the sequence of cosines of longitude for all nodes.

        **Example:**

        >>> r(GeoGrid.SmallTestGrid().cos_lon()[:2])
        array([ 0.999 , 0.9962])

        :rtype: 1D Numpy array [index]
        :return: the cosine of longitudes for all nodes.
        """
        return np.cos(self.lon_sequence() * np.pi / 180)

    def sin_lon(self):
        """
        Return the sequence of sines of longitude for all nodes.

        **Example:**

        >>> r(GeoGrid.SmallTestGrid().sin_lon()[:2])
        array([ 0.0436, 0.0872])

        :rtype: 1D Numpy array [index]
        :return: the sine of longitudes for all nodes.
        """
        return np.sin(self.lon_sequence() * np.pi / 180)

    def distance(self):
        """
        Calculate and return the standard distance matrix of the corresponding
        grid type

        :rtype: 2D Numpy array [index, index]
        :return: the distance matrix.
        """
        return self.angular_distance()

    @Cached.method(name="angular great circle distance")
    def angular_distance(self):
        """
        Calculate the angular great circle distance matrix.

        **No normalization applied anymore!** Return values are in the range
        0 to Pi.

        **Example:**

        >>> rr(GeoGrid.SmallTestGrid().angular_distance(), 2)
        [['0'    '0.1'  '0.19' '0.29' '0.39' '0.48']
         ['0.1'  '0'    '0.1'  '0.19' '0.29' '0.39']
         ['0.19' '0.1'  '0'    '0.1'  '0.19' '0.29']
         ['0.29' '0.19' '0.1'  '0'    '0.1'  '0.19']
         ['0.39' '0.29' '0.19' '0.1'  '0'    '0.1']
         ['0.48' '0.39' '0.29' '0.19' '0.1'  '0']]

        :rtype: 2D Numpy array [index, index]
        :return: the angular great circle distance matrix.
        """
        #  Get number of nodes
        N = self.N
        #  Initialize cython cof of angular distance matrix
        cosangdist = np.zeros((N, N), dtype=FIELD)
        _calculate_angular_distance(
            #  Get sequences of cosLat, sinLat, cosLon and sinLon for all nodes
            to_cy(self.cos_lat(), FIELD),
            to_cy(self.sin_lat(), FIELD),
            to_cy(self.cos_lon(), FIELD),
            to_cy(self.sin_lon(), FIELD),
            cosangdist, N)
        return np.arccos(cosangdist)

    def boundaries(self):
        """
        Return the spatio-temporal grid boundaries.

        Structure of the returned dictionary:
          - boundaries = {"time_min": self._boundaries["time_min"],
                          "time_max": self._boundaries["time_max"],
                          "lat_min": self._boundaries["space_min"][0],
                          "lat_max": self._boundaries["space_max"][1],
                          "lon_min": self._boundaries["space_min"][0],
                          "lon_max": self._boundaries["space_max"][1]}

        :rtype: dictionary
        :return: the spatio-temporal grid boundaries.
        """
        boundaries = {"time_min": self._boundaries["time_min"],
                      "time_max": self._boundaries["time_max"],
                      "lat_min": self._boundaries["space_min"][0],
                      "lat_max": self._boundaries["space_max"][0],
                      "lon_min": self._boundaries["space_min"][1],
                      "lon_max": self._boundaries["space_max"][1]}
        return boundaries

    def print_boundaries(self):
        """
        Pretty print the spatio-temporal grid boundaries.

        **Example:**

        >>> print(GeoGrid.SmallTestGrid().print_boundaries())
                 time     lat     lon
           min    0.0    0.00    2.50
           max    9.0   25.00   15.00

        :rtype: string
        :return: printable string for the spatio-temporal grid boundaries
        """
        return (
            "         time     lat     lon"
            "\n   min {time_min:6.1f} {lat_min: 7.2f} {lon_min: 7.2f}"
            "\n   max {time_max:6.1f} {lat_max: 7.2f} {lon_max: 7.2f}"
        ).format(**self.boundaries())

    def grid(self):
        """
        Return the grid's spatio-temporal sampling points.

        Structure of the returned dictionary:
          - grid = {"time": self._grid["time"],
                    "lat": self._grid["space"][0],
                    "lon": self._grid["space"][1]}

        **Examples:**

        >>> Grid.SmallTestGrid().grid()["space"][0]
        array([  0.,   5.,  10.,  15.,  20.,  25.], dtype=float32)
        >>> Grid.SmallTestGrid().grid()["space"][0][5]
        15.0

        :rtype: dictionary
        :return: the grid's spatio-temporal sampling points.
        """
        grid = {"time": self._grid["time"],
                "lat": self._grid["space"][0],
                "lon": self._grid["space"][1]}
        return grid

    #
    #  Methods for selecting regions
    #

    def region_indices(self, region):
        """
        Returns a boolean array of nodes with True values when the node
        is inside the region.

        **Example:**

        >>> GeoGrid.SmallTestGrid().region_indices(
        ...     np.array([0.,0.,0.,11.,11.,11.,11.,0.])).astype(int)
        array([0, 1, 1, 0, 0, 0])

        :type region: 1D Numpy array [n_polygon_nodes]
        :arg region: array of lon, lat, lon, lat, ...
                     [-80.2, 5., -82.4, 5.3, ...] as copied from Google Earth
                     Polygon file
        :rtype: 1D bool array [index]
        :return: bool array with True for nodes inside region
        """
        # Reshape Google Earth array  into (n,2) array
        remapped_region = region.reshape(len(region)//2, 2)
        # Remap from East-West to 360 degree map if the longitudes are [0, 360]
        if self._grid["space"][1].min() >= 0:
            remapped_region[remapped_region[:, 0] < 0, 0] = \
                360 + remapped_region[remapped_region[:, 0] < 0, 0]

        lat_lon_map = np.column_stack((self._grid["space"][1],
                                       self._grid["space"][0]))

        return path.Path(remapped_region).contains_points(lat_lon_map)

    @staticmethod
    def region(name):
        """Return some standard regions."""
        if name == 'ENSO':
            return np.array([-79.28273150749884, -10.49311965331937,
                             -79.29849791038734, 10.12527300655218,
                             -174.9221853596061, 10.07293121423917,
                             -174.8362810586096, -10.46407198776264,
                             -80.13229308153623, -10.36724072894785,
                             -79.28273150749884, -10.49311965331937])
        elif name == 'NINO34':
            return np.array([-118.6402427933005, 7.019906838300821,
                             -171.0067408177714, 6.215022481004243,
                             -171.0364908514962, -5.768616252424354,
                             -119.245702264066, -5.836385150138187,
                             -118.6402427933005, 7.019906838300821])
        else:
            return None



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Provides a mix-in class that manages an LRU cache of derived quantities
at the instance level, with declared dependencies on mutable attributes.
"""

from abc import ABC, abstractmethod
from functools import lru_cache, wraps
from inspect import getmembers, ismethod
from typing import Tuple, Optional
from collections.abc import Hashable


class Cached(ABC):
    """
    Mix-in class which manages an LRU cache for subclass methods that
    constitute derived quantities, based on `@functools.lru_cache()`.
    The cache is populated by calling decorated methods (with new arguments),
    has a bounded number of slots per method, and can be cleared manually.

    A subclasses should:
      - decorate derived quantity methods with `@Cached.method()`,
      - provide a method `__cache_state__() -> Tuple[Hashable,...]`, which is
        used by `Cached` to define the `__eq__()` and `__hash__()` methods
        used by `@functools.lru_cache()`, and
      - optionally provide a similar method to list owned `Cached` instances:
        `__rec_cache_state__() -> Tuple[object,...]`.

    Class attributes which affect subsequently *defined* subclasses:
      - cache_enable: toggles caching globally
      - lru_params:   sets `@functools.lru_cache()` parameters

    NOTE:
    The intended behaviour is specified by `tests/test_core/test_cache.py`.
    """
    cache_enable = True
    lru_params = {"maxsize": 32, "typed": True}

    @abstractmethod
    def __cache_state__(self) -> Tuple[Hashable, ...]:
        """
        Hashable tuple of mutable object attributes, which will determine the
        instance identity for ALL cached method lookups in this class,
        *in addition* to the built-in object `id()`. Returning an empty tuple
        amounts to declaring the object immutable in general. Mutable
        dependencies that are specific to a method should instead be declared
        via `@Cached.method(attrs=(...))`.

        NOTE:
        A subclass is responsible for the consistency and cost of this state
        descriptor. For example, hashing a large array attribute may be
        circumvented by declaring it as a property, with a custom setter method
        that increments a dedicated mutation counter.
        """

    def __rec_cache_state__(self) -> Tuple[object, ...]:
        """
        Similar to `__cache_state__()`, but lists attributes which are
        themselves instances of `Cached`. Empty by default.
        """
        return ()

    def __eq__(self, other):
        if self is not other:
            return False
        else:
            s, t = (o.__cache_state__() for o in [self, other])
            S, T = (o.__rec_cache_state__() for o in [self, other])
            assert all(isinstance(o, tuple) for o in [s, t])
            assert all(isinstance(o, tuple) for o in [S, T])
            assert len(s) == len(t) and len(S) == len(T)
            return s == t and all(s_ == t_ for s_, t_ in zip(S, T))

    def __hash__(self):
        s = self.__cache_state__()
        S = self.__rec_cache_state__()
        assert isinstance(s, tuple) and isinstance(S, tuple)
        return hash(sum(
          (((None,) if t is None else t.__cache_state__()) for t in S),
          start=(id(self),) + s))

    @classmethod
    def method(cls, name: Optional[str] = None,
               attrs: Optional[Tuple[str, ...]] = None):
        """
        Caching decorator based on `@functools.lru_cache()`.

        Cache entries for decorated methods are indexed by the combination of:
          - the object `id()`,
          - the object-level mutable instance attributes as declared by the
            subclass method `__cache_state__()`,
          - the object-level mutable instance attributes as declared by the
            optional subclass method `__rec_cache_state__()`,
          - the method-level mutable instance attributes as declared by the
            optional decorator argument `attrs`, and
          - the argument pattern at the call site, including the ordering of
            named arguments.

        The decorated method provides several attributes of its own, as defined
        by `@functools.lru_cache()`, including:
          - `cache_clear()`: delete this method cache for ALL `cls` instances
          - `__wrapped__`: the original method

        :arg name: Optionally print a message at the first method invocation.
        :arg attrs: Optionally declare attribute names as dependencies.

        NOTE:
        The same reasoning about consistency and cost applies to the `attrs`
        argument as to the `__cache_state__()` method.
        """
        # evaluated at decorator instantiation
        assert attrs is None or (
            isinstance(attrs, tuple) and len(attrs) > 0
            and all(isinstance(a, str) for a in attrs))
        assert name is None or isinstance(name, str)

        def wrapper(f):
            # evaluated at decorator application (method definition)
            if not cls.cache_enable:
                return f
            else:
                def uncached(self, *args, **kwargs):
                    # evaluated at uncached method invocation
                    if (
                      name is not None and
                      getattr(self, "silence_level", 0) <= 1):
                        print(f"Calculating {name}...")
                    return f(self, *args, **kwargs)

                if attrs is None:
                    # leave call signature unchanged
                    wrapped = lru_cache(**cls.lru_params)(uncached)
                else:
                    # present `attrs` as arguments for cache lookup
                    @lru_cache(**cls.lru_params)
                    def cached(self, *args, **kwargs):
                        return uncached(self, *args[len(attrs):], **kwargs)

                    @wraps(cached, assigned=(
                        'cache_info', 'cache_clear', '__wrapped__'))
                    def wrapped(self, *args, **kwargs):
                        _attrs = (getattr(self, a) for a in attrs)
                        return cached(self, *_attrs, *args, **kwargs)

                # fully decorated method
                return wraps(f)(wrapped)
        return wrapper

    def cache_clear(self, prefix: Optional[str] = None):
        """
        *Delete* all method caches for ALL instances of `self.__class__`,
        and recursively for owned `Cached` instances. This is simply a loop
        over the `cache_clear()` methods of all cached methods.

        :arg prefix: Optionally restrict the deleted caches by method name.

        NOTE:
        Instead, *invalidating* method caches for a SINGLE instance is achieved
        by modifying the relevant subset of mutable attributes, see
        `@Cached.method()`.
        """
        for n, m in getmembers(self, predicate=ismethod):
            if prefix is None or n.startswith(prefix):
                if all(hasattr(m, p) for p in ["cache_clear", "__wrapped__"]):
                    m.cache_clear()
        for c in self.__rec_cache_state__():
            c.cache_clear(prefix=prefix)



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"
"""
Simple tests for the Grid class.
"""
import numpy as np

from pyunicorn.core.grid import Grid


def test_RegularGrid():
    res = Grid.RegularGrid(time_seq=np.arange(2),
                           space_grid=np.array([[0., 5.], [1., 2.]]),
                           silence_level=2).sequence(0)
    exp = np.array([0., 0., 5., 5.], dtype=np.float32)
    assert np.allclose(res, exp, atol=1e-04)

    res = Grid.RegularGrid(time_seq=np.arange(2),
                           space_grid=np.array([[0., 5.], [1., 2.]]),
                           silence_level=2).sequence(1)
    exp = np.array([1., 2., 1., 2.], dtype=np.float32)
    assert np.allclose(res, exp, atol=1e-04)


def test_coord_sequence_from_rect_grid():
    res = Grid.coord_sequence_from_rect_grid(space_grid=np.array([[0., 5.],
                                                                  [1., 2.]]))
    exp = (np.array([0., 0., 5., 5.]), np.array([1., 2., 1., 2.]))
    assert np.allclose(res, exp, atol=1e-04)


def test_sequence():
    res = Grid.SmallTestGrid().sequence(0)
    exp = np.array([0., 5., 10., 15., 20., 25.], dtype=np.float32)
    assert np.allclose(res, exp, atol=1e-04)


def test_node_number():
    res = Grid.SmallTestGrid().node_number(x=(14., 9.))
    exp = 3
    assert res == exp


def test_node_coordinates():
    res = Grid.SmallTestGrid().node_coordinates(3)
    exp = (15.0, 10.0)
    assert np.allclose(res, exp, atol=1e-04)


def test_grid():
    res = Grid.SmallTestGrid().grid()["space"][0]
    exp = np.array([0., 5., 10., 15., 20., 25.], dtype=np.float32)
    assert np.allclose(res, exp, atol=1e-04)

    res = Grid.SmallTestGrid().grid()["space"][1]
    exp = np.array([2.5, 5., 7.5, 10., 12.5, 15.], dtype=np.float32)
    assert np.allclose(res, exp, atol=1e-04)


def test_grid_size():
    res = Grid.SmallTestGrid().print_grid_size()
    exp = '     space    time\n         6      10'
    assert res == exp


def test_geometric_distance_distribution():
    res = Grid.SmallTestGrid().geometric_distance_distribution(3)[0]
    exp = np.array([0.3333, 0.4667, 0.2])
    assert np.allclose(res, exp, atol=1e-04)

    res = Grid.SmallTestGrid().geometric_distance_distribution(3)[1]
    exp = np.array([0., 9.317, 18.6339, 27.9509])
    assert np.allclose(res, exp, atol=1e-04)


def test_euclidean_distance():
    res = Grid.SmallTestGrid().euclidean_distance().round(2)
    exp = np.array([[0., 5.59, 11.18, 16.77, 22.36, 27.95],
                    [5.59, 0., 5.59, 11.18, 16.77, 22.36],
                    [11.18, 5.59, 0., 5.59, 11.18, 16.77],
                    [16.77, 11.18, 5.59, 0., 5.59, 11.18],
                    [22.36, 16.77, 11.18, 5.59, 0., 5.59],
                    [27.95, 22.36, 16.77, 11.18, 5.59, 0.]], dtype=np.float32)
    assert (res == exp).all()



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"
"""
Simple tests for the GeoNetwork class.
"""
import numpy as np

from pyunicorn.core.geo_network import GeoNetwork
from pyunicorn.core.spatial_network import SpatialNetwork
from pyunicorn.core.geo_grid import GeoGrid


def test_ErdosRenyi(capsys):
    print(GeoNetwork.Model("ErdosRenyi", grid=GeoGrid.SmallTestGrid(),
                           n_nodes=6, n_links=5))
    out, err = capsys.readouterr()
    out_ref = "Generating Erdos-Renyi random graph with 6 nodes and 5 " + \
              "links...\nSetting area weights according to type surface " + \
              "...\nGeoNetwork:\nSpatialNetwork:\n" + \
              "Network: undirected, 6 nodes, 5 links, " + \
              "link density 0.333.\nGeographical boundaries:\n" + \
              "         time     lat     lon\n" + \
              "   min    0.0    0.00    2.50\n" + \
              "   max    9.0   25.00   15.00\n"
    assert out == out_ref


def test_ConfigurationModel():
    n = 0
    while n != 7:
        net = GeoNetwork.Model("Configuration", grid=GeoGrid.SmallTestGrid(),
                               degrees=GeoNetwork.SmallTestNetwork().degree(),
                               silence_level=2)
        n = net.n_links
    res = net.link_density
    exp = 0.46666667
    assert np.allclose(res, exp, atol=1e-04)


def test_randomly_rewire_geomodel_I():
    net = GeoNetwork.SmallTestNetwork()
    net.randomly_rewire_geomodel_I(distance_matrix=net.grid.distance(),
                                   iterations=100, inaccuracy=1.0)
    res = net.degree()
    exp = np.array([3, 3, 2, 2, 3, 1])
    assert (res == exp).all()


def test_set_random_links_by_distance():
    net = GeoNetwork.SmallTestNetwork()
    while net.n_links != 5:
        net.set_random_links_by_distance(a=0., b=-4.)
    res = net.n_links
    exp = 5
    assert res == exp


def test_geographical_distribution():
    net = GeoNetwork.SmallTestNetwork()
    res = net.geographical_distribution(sequence=net.degree(), n_bins=3)[0]
    exp = np.array([0.15645071, 0.33674395, 0.50680541])
    assert np.allclose(res, exp, atol=1e-04)


def test_geographical_cumulative_distribution():
    net = GeoNetwork.SmallTestNetwork()
    res = net.geographical_cumulative_distribution(sequence=net.degree(),
                                                   n_bins=3)[0]
    exp = np.array([1.00000007, 0.84354936, 0.50680541])
    assert np.allclose(res, exp, atol=1e-04)


def test_link_distance_distribution():
    """Check consistency of link distance distribution on spherical grid with
    distribution on euclidean grid"""
    net = SpatialNetwork.SmallTestNetwork()
    geo_net = GeoNetwork.SmallTestNetwork()

    res = geo_net.link_distance_distribution(n_bins=4,
                                             geometry_corrected=False,
                                             grid_type="spherical")[0]
    exp = net.link_distance_distribution(n_bins=4, geometry_corrected=False)[0]
    assert np.allclose(res, exp, atol=1e-04)

    res = geo_net.link_distance_distribution(n_bins=4, geometry_corrected=True,
                                             grid_type="spherical")[0]
    exp = net.link_distance_distribution(n_bins=4, geometry_corrected=True)[0]
    assert np.allclose(res, exp, atol=1e-04)


def test_area_weighted_connectivity():
    res = GeoNetwork.SmallTestNetwork().area_weighted_connectivity()
    exp = np.array([0.48540673, 0.4989577, 0.33418113, 0.34459165, 0.51459336,
                    0.17262428], dtype=np.float32)
    assert np.allclose(res, exp, atol=1e-04)


def test_inarea_weighted_connectivity():
    res = GeoNetwork.SmallTestNetwork().inarea_weighted_connectivity()
    exp = np.array([0.48540673, 0.4989577, 0.33418113, 0.34459165, 0.51459336,
                    0.17262428], dtype=np.float32)
    assert np.allclose(res, exp, atol=1e-04)


def test_outarea_weighted_connectivity():
    res = GeoNetwork.SmallTestNetwork().outarea_weighted_connectivity()
    exp = np.array([0.4854067, 0.4989577, 0.33418113, 0.34459165, 0.51459336,
                    0.17262428], dtype=np.float32)
    assert np.allclose(res, exp, atol=1e-04)


def test_area_weighted_connectivity_distribution():
    net = GeoNetwork.SmallTestNetwork()
    res = net.area_weighted_connectivity_distribution(n_bins=4)[0]
    exp = np.array([0.15645071, 0.33674395, 0.34459165, 0.16221375])
    assert np.allclose(res, exp, atol=1e-04)


def test_inarea_weighted_connectivity_distribution():
    net = GeoNetwork.SmallTestNetwork()
    res = net.inarea_weighted_connectivity_distribution(n_bins=4)[0]
    exp = np.array([0.15645071, 0.33674395, 0.34459165, 0.16221375])
    assert np.allclose(res, exp, atol=1e-04)


def test_outarea_weighted_connectivity_distribution():
    net = GeoNetwork.SmallTestNetwork()
    res = net.outarea_weighted_connectivity_distribution(n_bins=4)[0]
    exp = np.array([0.15645071, 0.33674395, 0.34459165, 0.16221375])
    assert np.allclose(res, exp, atol=1e-04)


def test_area_weighted_connectivity_cumulative_distribution():
    net = GeoNetwork.SmallTestNetwork()
    res = net.area_weighted_connectivity_cumulative_distribution(n_bins=4)[0]
    exp = np.array([1.00000007, 0.84354936, 0.50680541, 0.16221375])
    assert np.allclose(res, exp, atol=1e-04)


def test_inarea_weighted_connectivity_cumulative_distribution():
    net = GeoNetwork.SmallTestNetwork()
    res = net.inarea_weighted_connectivity_cumulative_distribution(n_bins=4)[0]
    exp = np.array([1.00000007, 0.84354936, 0.50680541, 0.16221375])
    assert np.allclose(res, exp, atol=1e-04)


def test_outarea_weighted_connectivity_cumulative_distribution():
    net = GeoNetwork.SmallTestNetwork()
    res = net.outarea_weighted_connectivity_cumulative_distribution(
        n_bins=4)[0]
    exp = np.array([1.00000007, 0.84354936, 0.50680541, 0.16221375])
    assert np.allclose(res, exp, atol=1e-04)


def test_average_neighbor_area_weighted_connectivity():
    net = GeoNetwork.SmallTestNetwork()
    res = net.average_neighbor_area_weighted_connectivity()
    exp = np.array([0.3439364, 0.39778873, 0.5067755, 0.4921822, 0.4395152,
                    0.48540673], dtype=np.float32)
    assert np.allclose(res, exp, atol=1e-04)


def test_max_neighbor_area_weighted_connectivity():
    net = GeoNetwork.SmallTestNetwork()
    res = net.max_neighbor_area_weighted_connectivity()
    exp = np.array([0.51459336, 0.51459336, 0.51459336, 0.49895769, 0.49895769,
                    0.48540673])
    assert np.allclose(res, exp, atol=1e-04)


def test_average_link_distance():
    net = GeoNetwork.SmallTestNetwork()

    res = net.average_link_distance(geometry_corrected=False)
    exp = np.array([0.3884563, 0.19434186, 0.14557932, 0.24325928, 0.29118925,
                    0.48467883])
    assert np.allclose(res, exp, atol=1e-04)

    res = net.average_link_distance(geometry_corrected=True)[:-1]
    exp = np.array([1.5987561, 1.0920637, 1.00011046, 1.67081917, 1.63841055])
    assert np.allclose(res, exp, atol=1e-04)


def test_inaverage_link_distance():
    net = GeoNetwork.SmallTestNetwork()
    res = net.inaverage_link_distance(geometry_corrected=False)
    exp = np.array([0.3884563, 0.19434186, 0.14557932, 0.24325928, 0.29118925,
                    0.48467883])
    assert np.allclose(res, exp, atol=1e-04)


def test_outaverage_link_distance():
    net = GeoNetwork.SmallTestNetwork()
    res = net.outaverage_link_distance(geometry_corrected=False)
    exp = np.array([0.3884563, 0.19434186, 0.14557932, 0.24325928, 0.29118925,
                    0.48467883])
    assert np.allclose(res, exp, atol=1e-04)


def test_total_link_distance():
    net = GeoNetwork.SmallTestNetwork()
    res = net.total_link_distance(geometry_corrected=False)
    exp = np.array([0.1885593, 0.09696837, 0.04864986, 0.08382512, 0.14984406,
                    0.08366733])
    assert np.allclose(res, exp, atol=1e-04)


def test_intotal_link_distance():
    net = GeoNetwork.SmallTestNetwork()
    res = net.intotal_link_distance(geometry_corrected=False)
    exp = np.array([0.1885593, 0.09696837, 0.04864986, 0.08382512, 0.14984406,
                    0.08366733])
    assert np.allclose(res, exp, atol=1e-04)


def test_outtotal_link_distance():
    net = GeoNetwork.SmallTestNetwork()
    res = net.outtotal_link_distance(geometry_corrected=False)
    exp = np.array([0.18855929, 0.09696837, 0.04864986, 0.08382512, 0.14984406,
                    0.08366733])
    assert np.allclose(res, exp, atol=1e-04)


def test_connectivity_weighted_distance():
    net = GeoNetwork.SmallTestNetwork()
    res = net.connectivity_weighted_distance()
    exp = np.array([0.0625227, 0.03207134, 0.02408994, 0.04192858, 0.0500332,
                    0.08366733])
    assert np.allclose(res, exp, atol=1e-04)


def test_inconnectivity_weighted_distance():
    net = GeoNetwork.SmallTestNetwork()
    res = net.inconnectivity_weighted_distance()
    exp = np.array([0.0625227, 0.03207134, 0.02408994, 0.04192858, 0.0500332,
                    0.08366733])
    assert np.allclose(res, exp, atol=1e-04)


def test_outconnectivity_weighted_distance():
    net = GeoNetwork.SmallTestNetwork()
    res = net.outconnectivity_weighted_distance()
    exp = np.array([0.0625227, 0.03207134, 0.02408994, 0.04192858, 0.0500332,
                    0.08366733])
    assert np.allclose(res, exp, atol=1e-04)


def test_max_link_distance():
    res = GeoNetwork.SmallTestNetwork().max_link_distance()
    exp = np.array([0.48467883, 0.2911402, 0.19376467, 0.2920272, 0.38866287,
                    0.48467883], dtype=np.float32)
    assert np.allclose(res, exp, atol=1e-04)


def test_average_distance_weighted_path_length():
    res = GeoNetwork.SmallTestNetwork().average_distance_weighted_path_length()
    exp = 0.49846491
    assert np.allclose(res, exp, atol=1e-04)


def test_distance_weighted_closeness():
    res = GeoNetwork.SmallTestNetwork().distance_weighted_closeness()
    exp = np.array([2.23782229, 2.45008978, 2.23956348, 2.45008978, 2.2396005,
                    1.19817005])
    assert np.allclose(res, exp, atol=1e-04)


def test_local_distance_weighted_vulnerability():
    res = GeoNetwork.SmallTestNetwork().local_distance_weighted_vulnerability()
    exp = np.array([0.03252519, 0.3136603, 0.20562718, 0.02799094, -0.02828809,
                    -0.28798557])
    assert np.allclose(res, exp, atol=1e-04)


def test_local_geographical_clustering():
    res = GeoNetwork.SmallTestNetwork().local_geographical_clustering()
    exp = np.array([0., 0.0998, 0.1489, 0., 0.2842, 0.])
    assert np.allclose(res, exp, atol=1e-04)



import numpy
from pyunicorn import Network, mpi

offset = 10
n_max = 1000
s = 0
n = mpi.rank + offset
while n <= n_max + offset:
    s += Network.BarabasiAlbert(n_nodes=n).global_clustering()
    n += mpi.size

numpy.save("s"+str(mpi.rank), s)



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"
"""
Simple tests for the SpatialNetwork class.
"""
import numpy as np

from pyunicorn.core.geo_network import SpatialNetwork


def test_randomly_rewire_geomodel_I():
    net = SpatialNetwork.SmallTestNetwork()
    net.randomly_rewire_geomodel_I(distance_matrix=net.grid.distance(),
                                   iterations=100, inaccuracy=100)
    res = net.degree()
    exp = np.array([3, 3, 2, 2, 3, 1])
    assert (res == exp).all()


def test_set_random_links_by_distance():
    net = SpatialNetwork.SmallTestNetwork()
    while net.n_links != 5:
        net.set_random_links_by_distance(a=0., b=-0.04)
    res = net.n_links
    exp = 5
    assert res == exp


def test_link_distance_distribution():
    net = SpatialNetwork.SmallTestNetwork()

    res = net.link_distance_distribution(n_bins=4, geometry_corrected=False)[0]
    exp = np.array([0.14285714, 0.28571429, 0.28571429, 0.28571429])
    assert np.allclose(res, exp, atol=1e-04)

    res = net.link_distance_distribution(n_bins=4, geometry_corrected=True)[0]
    exp = np.array([0.09836066, 0.24590164, 0.32786885, 0.32786885])
    assert np.allclose(res, exp, atol=1e-04)


def test_average_link_distance():
    net = SpatialNetwork.SmallTestNetwork()

    res = net.average_link_distance(geometry_corrected=False)
    exp = np.array([22.36067963, 11.18033981, 8.38525486, 13.97542477,
                    16.77050908, 27.95084953])
    assert np.allclose(res, exp, atol=1e-04)

    res = net.average_link_distance(geometry_corrected=True)[:-1]
    exp = np.array([1.6, 1.09090909, 1., 1.66666667, 1.63636357])
    assert np.allclose(res, exp, atol=1e-04)


def test_inaverage_link_distance():
    net = SpatialNetwork.SmallTestNetwork()
    res = net.inaverage_link_distance(geometry_corrected=False)
    exp = np.array([22.36067963, 11.18033981, 8.38525486, 13.97542477,
                    16.77050908, 27.95084953])
    assert np.allclose(res, exp, atol=1e-04)


def test_outaverage_link_distance():
    net = SpatialNetwork.SmallTestNetwork()
    res = net.outaverage_link_distance(geometry_corrected=False)
    exp = np.array([22.36067963, 11.18033981, 8.38525486, 13.97542477,
                    16.77050908, 27.95084953])
    assert np.allclose(res, exp, atol=1e-04)


def test_max_link_distance():
    res = SpatialNetwork.SmallTestNetwork().max_link_distance()
    exp = np.array([27.95085, 16.77051, 11.18034, 16.77051, 22.36068,
                    27.95085], dtype=np.float32)
    assert np.allclose(res, exp, atol=1e-04)


def test_average_distance_weighted_path_length():
    res = SpatialNetwork.SmallTestNetwork(). \
            average_distance_weighted_path_length()
    exp = 28.69620552062988
    assert np.allclose(res, exp, atol=1e-04)


def test_distance_weighted_closeness():
    res = SpatialNetwork.SmallTestNetwork().distance_weighted_closeness()
    exp = np.array([0.03888814, 0.04259177, 0.03888814, 0.04259177, 0.03888814,
                    0.02080063])
    assert np.allclose(res, exp, atol=1e-04)


def test_local_distance_weighted_vulnerability():
    res = SpatialNetwork.SmallTestNetwork(). \
            local_distance_weighted_vulnerability()
    exp = np.array([0.03233506, 0.31442454, 0.20580213, 0.02843829,
                    -0.02929477, -0.2883446])
    assert np.allclose(res, exp, atol=1e-04)



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Provides classes for the analysis of dynamical systems and time series based
on recurrence plots, including measures of recurrence quantification
analysis (RQA) and recurrence network analysis.
"""

from typing import Tuple
from collections.abc import Hashable

import numpy as np

from ..core.cache import Cached
from .recurrence_plot import RecurrencePlot
from ..core._ext.types import to_cy, DFIELD
from ._ext.numerics import _manhattan_distance_matrix_crp, \
    _euclidean_distance_matrix_crp, _supremum_distance_matrix_crp


class CrossRecurrencePlot(RecurrencePlot):
    """
    Class CrossRecurrencePlot for generating and quantitatively analyzing
    :index:`cross recurrence plots <single: cross recurrence plot>`.

    The CrossRecurrencePlot class supports the construction of cross recurrence
    plots from two multi-dimensional time series, optionally using embedding.
    Currently, manhattan, euclidean and supremum norms are provided for
    measuring distances in phase space.

    Methods for calculating commonly used measures of :index:`recurrence
    quantification analysis <pair: RQA; cross recurrence plot>` (RQA) are
    provided, e.g., determinism, maximum diagonal line length and laminarity.
    The definitions of these measures together with a review of the theory and
    applications of cross recurrence plots are given in [Marwan2007]_.

    **Examples:**

     - Create an instance of CrossRecurrencePlot from time series x and y with
       a fixed recurrence threshold and without embedding::

           CrossRecurrencePlot(x, y, threshold=0.1)

     - Create an instance of CrossRecurrencePlot at a fixed recurrence rate and
       using time delay embedding::

           CrossRecurrencePlot(x, y, dim=3, tau=2,
                               recurrence_rate=0.05).recurrence_rate()
    """

    #
    #  Internal methods
    #

    def __init__(self, x, y, metric="supremum", normalize=False,
                 sparse_rqa=False, silence_level=0, **kwds):
        """
        Initialize an instance of CrossRecurrencePlot.

        .. note::
           For a cross recurrence plot, time series x and y generally do
           **not** need to have the same length!

        Either recurrence threshold ``threshold`` or recurrence rate
        ``recurrence_rate`` have to be given as keyword arguments.

        Embedding is only supported for scalar time series. If embedding
        dimension ``dim`` and delay ``tau`` are **both** given as keyword
        arguments, embedding is applied. Multidimensional time series are
        processed as is by default. The same delay embedding is applied to both
        time series x and y.

        :type x: 2D array (time, dimension)
        :arg x: One of the time series to be analyzed, can be scalar or
            multi-dimensional.
        :type y: 2D array (time, dimension)
        :arg y: One of the time series to be analyzed, can be scalar or
            multi-dimensional.
        :arg str metric: The metric for measuring distances in phase space
            ("manhattan", "euclidean", "supremum").
        :arg bool normalize: Decide whether to normalize both time series to
            zero mean and unit standard deviation.
        :arg number silence_level: Inverse level of verbosity of the object.
        :arg number threshold: The recurrence threshold keyword for generating
            the cross recurrence plot using a fixed threshold.
        :arg number recurrence_rate: The recurrence rate keyword for generating
            the cross recurrence plot using a fixed recurrence rate.
        :arg number dim: The embedding dimension.
        :arg number tau: The embedding delay.
        """
        threshold = kwds.get("threshold")
        recurrence_rate = kwds.get("recurrence_rate")

        RecurrencePlot.__init__(
            self, np.empty((2, 0)), metric=metric, normalize=normalize,
            sparse_rqa=sparse_rqa, silence_level=silence_level,
            skip_recurrence=True)

        self.CR = None
        """The cross recurrence matrix."""
        self.N = 0
        """The length of the embedded time series x."""
        self.M = 0
        """The length of the embedded time series y."""

        #  Store time series
        self.x = x.copy()
        """The time series x."""
        self.y = y.copy()
        """The time series y."""

        #  Reshape time series
        self.x.shape = (self.x.shape[0], -1)
        self.y.shape = (self.y.shape[0], -1)

        #  Normalize time series
        if normalize:
            self.normalize_time_series(self.x)
            self.normalize_time_series(self.y)

        #  Get embedding dimension and delay from **kwds
        dim = kwds.get("dim")
        tau = kwds.get("tau")

        self._mut_embedding: int = 0
        if (dim is not None) and (tau is not None):
            #  Embed the time series
            self.x_embedded = self.embed_time_series(self.x, dim, tau)
            self.y_embedded = self.embed_time_series(self.y, dim, tau)
        else:
            self.x_embedded = self.x
            self.y_embedded = self.y

        #  construct recurrence plot accordingly to threshold / recurrence rate
        if threshold is not None:
            #  Calculate the recurrence matrix R using the radius of
            #  neighborhood threshold
            CrossRecurrencePlot.set_fixed_threshold(self, threshold)
        elif recurrence_rate is not None:
            #  Calculate the recurrence matrix R using a fixed recurrence rate
            CrossRecurrencePlot.\
                set_fixed_recurrence_rate(self, recurrence_rate)
        else:
            raise NameError("Please give either threshold or recurrence_rate \
                            to construct the cross recurrence plot!")

    def __cache_state__(self) -> Tuple[Hashable, ...]:
        return (self._mut_embedding,)

    def __str__(self):
        """
        Returns a string representation.
        """
        return ("CrossRecurrencePlot: "
                f"time series shapes {self.x.shape}, {self.y.shape}.\n"
                f"Embedding dimension {self.dim if self.dim else 0}\n"
                f"Threshold {self.threshold}, {self.metric} metric")

    @property
    def x_embedded(self) -> np.ndarray:
        """
        The embedded time series x.
        """
        return self._x_embedded

    @x_embedded.setter
    def x_embedded(self, embedding: np.ndarray):
        self._x_embedded = to_cy(embedding, DFIELD)
        self._mut_embedding += 1

    @property
    def y_embedded(self) -> np.ndarray:
        """
        The embedded time series y.
        """
        return self._y_embedded

    @y_embedded.setter
    def y_embedded(self, embedding: np.ndarray):
        self._y_embedded = to_cy(embedding, DFIELD)
        self._mut_embedding += 1

    #
    #  Service methods
    #

    def recurrence_matrix(self):
        """
        Return the current cross recurrence matrix :math:`CR`.

        :rtype: 2D square Numpy array
        :return: the current cross recurrence matrix :math:`CR`.
        """
        return self.CR

    def distance_matrix(self, metric):
        """
        Return phase space cross distance matrix :math:`D` according to the
        chosen metric.

        :arg str metric: The metric for measuring distances in phase space
            ("manhattan", "euclidean", "supremum").
        :rtype: 2D square array
        :return: the phase space cross distance matrix :math:`D`
        """
        assert metric in self._known_metrics, f"unknown metric: {metric}"
        return getattr(self, f"{metric}_distance_matrix")()

    #
    #  Calculate recurrence plot
    #

    @Cached.method(name="the manhattan distance matrix")
    def manhattan_distance_matrix(self):
        """
        Return the manhattan distance matrix from two (embedded) time series.

        :type x_embedded: 2D Numpy array (time, embedding dimension)
        :arg x_embedded: The phase space trajectory x.
        :type y_embedded: 2D Numpy array (time, embedding dimension)
        :arg y_embedded: The phase space trajectory y.
        :rtype: 2D rectangular Numpy array
        :return: the manhattan distance matrix.
        """
        ntime_x = self.x_embedded.shape[0]
        ntime_y = self.y_embedded.shape[0]
        dim = self.x_embedded.shape[1]
        return _manhattan_distance_matrix_crp(ntime_x, ntime_y, dim,
                                              self.x_embedded, self.y_embedded)

    @Cached.method(name="the euclidean distance matrix")
    def euclidean_distance_matrix(self):
        """
        Return the euclidean distance matrix from two (embedded) time series.

        :rtype: 2D rectangular Numpy array
        :return: the euclidean distance matrix.
        """
        ntime_x = self.x_embedded.shape[0]
        ntime_y = self.y_embedded.shape[0]
        dim = self.x_embedded.shape[1]
        return _euclidean_distance_matrix_crp(ntime_x, ntime_y, dim,
                                              self.x_embedded, self.y_embedded)

    @Cached.method(name="the supremum distance matrix")
    def supremum_distance_matrix(self):
        """
        Return the supremum distance matrix from two (embedded) time series.

        :rtype: 2D rectangular Numpy array
        :return: the supremum distance matrix.
        """
        ntime_x = self.x_embedded.shape[0]
        ntime_y = self.y_embedded.shape[0]
        dim = self.x_embedded.shape[1]
        return _supremum_distance_matrix_crp(ntime_x, ntime_y, dim,
                                             self.x_embedded, self.y_embedded)

    def set_fixed_threshold(self, threshold):
        """
        Set the cross recurrence plot to a fixed threshold.

        Modifies / sets the class variables :attr:`CR`, :attr:`N` and :attr:`M`
        accordingly.

        :arg number threshold: The recurrence threshold.
        """
        if self.silence_level <= 1:
            print("Calculating cross recurrence plot at fixed threshold...")

        distance = self.distance_matrix(self.metric)
        (N, M) = distance.shape
        recurrence = np.zeros((N, M), dtype="int8")
        recurrence[distance < threshold] = 1
        self.CR = recurrence
        self.N = N
        self.M = M

    def set_fixed_recurrence_rate(self, recurrence_rate):
        """
        Set the cross recurrence plot to a fixed recurrence rate.

        Modifies / sets the class variables :attr:`CR`, :attr:`N` and :attr:`M`
        accordingly.

        :arg number recurrence_rate: The recurrence rate.
        """
        if self.silence_level <= 1:
            print("Calculating cross recurrence plot at "
                  "fixed recurrence rate...")

        distance = self.distance_matrix(self.metric)
        (N, M) = distance.shape
        threshold = self.threshold_from_recurrence_rate(distance,
                                                        recurrence_rate)
        recurrence = np.zeros((N, M), dtype="int8")
        recurrence[distance < threshold] = 1
        self.CR = recurrence
        self.N = N
        self.M = M

    #
    #  Extended RQA measures
    #

    def recurrence_rate(self):
        """
        Return cross recurrence rate.

        Alias to :meth:`cross_recurrence_rate`, since
        :meth:`RecurrencePlot.recurrence_rate` would give incorrect results
        here.

        :rtype: number (float)
        :return: the cross recurrence rate.
        """
        return self.cross_recurrence_rate()

    def cross_recurrence_rate(self):
        """
        Return cross recurrence rate.

        :rtype: number (float)
        :return: the cross recurrence rate.
        """
        return float(self.CR.sum()) / (self.N * self.M)

    def balance(self):
        """
        Return balance of the cross recurrence plot.

        Might be useful for detecting the direction of coupling between systems
        using cross recurrence analysis.
        """
        #  Get cross recurrence matrix
        CR = self.recurrence_matrix()

        #  Get sum of upper triangle of cross recurrence matrix, excluding the
        #  main diagonal
        upper_triangle_sum = np.triu(CR, k=1).sum()

        #  Get sum of lower triangle of cross recurrence matrix, excluding the
        #  main diagonal
        lower_triangle_sum = np.tril(CR, k=-1).sum()

        #  Return balance
        return (upper_triangle_sum - lower_triangle_sum) / \
            float(upper_triangle_sum + lower_triangle_sum)

    def diagline_dist(self):
        """Not implemented yet"""
        raise NotImplementedError(
            "Line distributions are not yet "
            "available for cross-recurrence plots")

    def vertline_dist(self):
        """Not implemented yet"""
        raise NotImplementedError(
            "Line distributions are not yet "
            "available for cross-recurrence plots")

    def white_vertline_dist(self):
        """Not implemented yet"""
        raise NotImplementedError(
            "Line distributions are not yet "
            "available for cross-recurrence plots")



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Provides classes for the analysis of dynamical systems and time series based
on recurrence plots, including measures of recurrence quantification
analysis (RQA) and recurrence network analysis.
"""

# array object and fast numerics
import numpy as np

from ..core import Network
from .recurrence_plot import RecurrencePlot


#
#  Class definitions
#

class RecurrenceNetwork(RecurrencePlot, Network):

    """
    Class RecurrenceNetwork for generating and quantitatively analyzing
    recurrence networks.

    More information on the theory and applications of recurrence networks can
    be found in [Marwan2009]_, [Donner2010b]_.

    Examples:

     - Create an instance of RecurrenceNetwork with a fixed recurrence
       threshold and without embedding::

           RecurrenceNetwork(time_series, threshold=0.1)

     - Create an instance of RecurrenceNetwork at a fixed (global) recurrence
       rate and using time delay embedding::

           RecurrenceNetwork(time_series, dim=3, tau=2,
                             recurrence_rate=0.05).recurrence_rate()
    """

    #
    #  Internal methods
    #

    def __init__(self, time_series, metric="supremum", normalize=False,
                 missing_values=False, silence_level=0, **kwds):
        """
        Initialize an instance of RecurrenceNetwork.

        Creates an embedding of the given time series, calculates a recurrence
        plot from the embedding and then creates a Network object from the
        recurrence plot, interpreting the recurrence matrix as the adjacency
        matrix of a complex network.

        Either recurrence threshold ``threshold``/``threshold_std``, recurrence
        rate ``recurrence_rate`` or local recurrence rate
        ``local_recurrence_rate`` have to be given as keyword arguments.

        Embedding is only supported for scalar time series. If embedding
        dimension ``dim`` and delay ``tau`` are **both** given as keyword
        arguments, embedding is applied. Multidimensional time series are
        processed as is by default.

        :type time_series: 2D array (time, dimension)
        :arg time_series: The time series to be analyzed, can be scalar or
            multi-dimensional.
        :arg str metric: The metric for measuring distances in phase space
            ("manhattan", "euclidean", "supremum").
        :arg bool normalize: Decide whether to normalize the time series to
            zero mean and unit standard deviation.
        :arg bool missing_values: Toggle special treatment of missing values in
            :attr:`.RecurrencePlot.time_series`.
        :arg number silence_level: Inverse level of verbosity of the object.
        :arg number threshold: The recurrence threshold keyword for generating
            the recurrence network using a fixed threshold.
        :arg number threshold_std: The recurrence threshold keyword for
            generating the recurrence plot using a fixed threshold in units of
            the time series' STD.
        :arg number recurrence_rate: The recurrence rate keyword for generating
            the recurrence network using a fixed recurrence rate.
        :arg number local_recurrence_rate: The local recurrence rate keyword
            for generating the recurrence plot using a fixed local recurrence
            rate (same number of recurrences for each state vector).
        :arg number adaptive_neighborhood_size: The adaptive neighborhood size
            parameter for generating recurrence plots based on the algorithm in
            [Xu2008]_.
        :arg number dim: The embedding dimension.
        :arg number tau: The embedding delay.
        :type node_weights: 1D array (time)
        :arg node_weights: The sequence of weights associated with each
            node for calculating n.s.i. network measures.
        """
        #  Initialize the underlying RecurrencePlot object
        RecurrencePlot.__init__(self, time_series=time_series, metric=metric,
                                normalize=normalize,
                                missing_values=missing_values,
                                silence_level=silence_level, **kwds)

        #  Set diagonal of R to zero to avoid self-loops in the recurrence
        #  network
        A = self.R.copy()
        A.flat[::self.N+1] = 0

        #  Get keywords
        local_recurrence_rate = kwds.get("local_recurrence_rate")
        node_weights = kwds.get("node_weights")

        #  Assign to each embedded state vector the weight of the earliest
        #  sample contained in it (too simple!)
        if node_weights is not None:
            node_weights = node_weights[:self.N]

        #  Remove state vectors containing missing values
        if missing_values:
            A = np.delete(A, np.where(self.missing_value_indices), axis=0)
            A = np.delete(A, np.where(self.missing_value_indices), axis=1)

        #  Create a Network object interpreting the recurrence matrix as the
        #  graph adjacency matrix.
        if local_recurrence_rate is not None:
            #  A recurrence network with fixed local recurrence rate (Eckmann
            #  definition of a recurrence plot) is directed by definition.
            Network.__init__(self, A, directed=True, node_weights=node_weights,
                             silence_level=silence_level)
        else:
            #  Create a Network object interpreting the recurrence matrix as
            #  the graph adjacency matrix. Recurrence networks are undirected
            #  by definition.
            Network.__init__(self, A, directed=False,
                             node_weights=node_weights,
                             silence_level=silence_level)

    def __str__(self):
        """
        Returns a string representation.
        """
        return (f"RecurrenceNetwork:\n"
                f"{RecurrencePlot.__str__(self)}\n"
                f"{Network.__str__(self)}")

    def clear_cache(self):
        """
        Clean up memory by deleting information that can be recalculated from
        basic data.

        Extends the clean up methods of the parent classes.
        """
        #  Call clean up of RecurrencePlot
        RecurrencePlot.clear_cache(self)
        #  Call clean up of Network
        Network.clear_cache(self)

    #
    #  Methods to handle recurrence networks
    #
    def set_fixed_threshold(self, threshold):
        """
        Create a recurrence network at a fixed threshold.

        :arg number threshold: The threshold.
        """
        #  Set fixed threshold on recurrence plot level
        RecurrencePlot.set_fixed_threshold(self, threshold)

        #  Set diagonal of R to zero to avoid self-loops in the recurrence
        #  network
        A = self.R.copy()
        A.flat[::self.N+1] = 0

        #  Create a Network object interpreting the recurrence matrix as the
        #  graph adjacency matrix. Recurrence networks are undirected by
        #  definition.
        Network.__init__(self, A, directed=False,
                         silence_level=self.silence_level)

    def set_fixed_threshold_std(self, threshold_std):
        """
        Set the recurrence network to a fixed threshold in units of the
        standard deviation of the time series.

        :arg number threshold_std: The recurrence threshold in units of the
            standard deviation of the time series.
        """
        #  Set fixed threshold in units of STD on recurrence plot level
        RecurrencePlot.set_fixed_threshold_std(self, threshold_std)

        #  Set diagonal of R to zero to avoid self-loops in the recurrence
        #  network
        A = self.R.copy()
        A.flat[::self.N+1] = 0

        #  Create a Network object interpreting the recurrence matrix as the
        #  graph adjacency matrix. Recurrence networks are undirected by
        #  definition.
        Network.__init__(self, A, directed=False,
                         silence_level=self.silence_level)

    def set_fixed_recurrence_rate(self, recurrence_rate):
        """
        Create a recurrence network at a fixed link density (recurrence rate).

        :arg number recurrence_rate: The link density / recurrence rate.
        """
        #  Set fixed recurrence rate on recurrence plot level
        RecurrencePlot.set_fixed_recurrence_rate(self, recurrence_rate)

        #  Set diagonal of R to zero to avoid self-loops in the recurrence
        #  network
        A = self.R.copy()
        A.flat[::self.N+1] = 0

        #  Create a Network object interpreting the recurrence matrix as the
        #  graph adjacency matrix. Recurrence networks are undirected by
        #  definition.
        Network.__init__(self, A, directed=False,
                         silence_level=self.silence_level)

    def set_fixed_local_recurrence_rate(self, local_recurrence_rate):
        """
        Create a recurrence network at a fixed local recurrence rate.

        This leads to a directed recurrence network with identical out-degree
        :math:`int(N * local_recurrence_rate)`, but variable in-degree. The
        associated recurrence plot coincides with the original Eckmann
        definition.

        :arg number local_recurrence_rate: The local recurrence rate.
        """
        #  Set fixed local recurrence rate on recurrence plot level
        RecurrencePlot.\
            set_fixed_local_recurrence_rate(self, local_recurrence_rate)

        #  Create a Network object interpreting the recurrence matrix as the
        #  graph adjacency matrix. Set diagonal of R to zero to avoid
        #  self-loops in the recurrence network
        A = self.R.copy()
        A.flat[::self.N+1] = 0

        #  A recurrence network with fixed local recurrence rate (Eckmann
        #  definition of a recurrence plot) is directed by definition.
        Network.__init__(self, A, directed=True,
                         silence_level=self.silence_level)

    def set_adaptive_neighborhood_size(self, adaptive_neighborhood_size,
                                       order=None):
        """
        Create a recurrence network using the :index:`adaptive neighborhood
        size <single: adaptive neighborhood size; recurrence network>`
        algorithm used in [Xu2008]_.

        The exact algorithm was deduced from private correspondence with the
        authors.  It leads to an undirected network with mean degree
        :math:`<k> = 2 * m`, where m is the adaptive_neighborhood_size.
        The degree :math:`k_v` of single nodes may vary, but :math:`k_v >= m`
        holds!

        :arg number adaptive_neighborhood_size: The number of adaptive nearest
            neighbors (recurrences) assigned to each state vector.
        :type order: 1D array (int32)
        :arg order: The indices of state vectors in the order desired for
            processing by the algorithm.
        """
        #  Set adaptive neighborhood size on recurrence plot level
        RecurrencePlot.set_adaptive_neighborhood_size(
            self, adaptive_neighborhood_size, order)

        #  Create a Network object interpreting the recurrence matrix as the
        #  graph adjacency matrix. Set diagonal of R to zero to avoid
        #  self-loops in the recurrence network
        A = self.R.copy()
        A.flat[::self.N+1] = 0

        #  A recurrence network with fixed local recurrence rate (Eckmann
        #  definition of a recurrence plot) is directed by definition.
        Network.__init__(self, A, directed=False,
                         silence_level=self.silence_level)

    def transitivity_dim_single_scale(self):
        """
        Return transitivity dimension for a single scale.

        The single scale transitivity dimension can be interpreted as a global
        measure of the dimensionality of the set of points underlying the
        recurrence network ([Donner2011b]_.). The scale is determined by the
        chosen recurrence threshold. Note that the maxima and minima of the
        single scale transitivity dimension when varying the scale give a
        more meaningful measure of dimensionality as is explained in
        [Donner2011b]_.

        **Attention:** currently only works correctly for supremum norm.

        :rtype: float
        :return: the single scale transitivity dimension.
        """
        return np.log(self.transitivity()) / np.log(3. / 4.)

    def local_clustering_dim_single_scale(self):
        """
        Return local clustering dimension for a single scale.

        The single scale local clustering dimension can be interpreted as a
        local measureof the dimensionality of the set of points underlying the
        recurrence network ([Donner2011b]_.). The scale is determined by the
        chosen recurrence threshold. Note that the maxima and minima of the
        single scale local clustering dimension when varying the scale give a
        more meaningful measure of dimensionality as is explained in
        [Donner2011b]_.

        **Attention:**
        currently only works correctly for supremum norm.

        :rtype: 1d numpy array [node] of float
        :return: the single scale transitivity dimension.
        """
        return np.log(self.local_clustering()) / np.log(3. / 4.)



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Provides classes for generating and analyzing complex climate networks.
"""

from typing import Tuple
from collections.abc import Hashable, Callable

import numpy as np
import igraph

from ..core.cache import Cached
from ..core import GeoNetwork, GeoGrid


class ClimateNetwork(GeoNetwork):

    """
    Encapsulates a similarity network embedded on a spherical surface.

    Particularly provides functionality to generate a complex network from the
    matrix of a similarity measure of time series.

    The analysis of climate time series based on similarity networks was first
    introduced in [Tsonis2004]_.
    """
    #
    #  Definitions of internal methods
    #

    def __init__(self, grid: GeoGrid, similarity_measure: np.ndarray,
                 threshold=None, link_density=None, non_local=False,
                 directed=False, node_weight_type="surface", silence_level=0):
        """
        Initialize an instance of :class:`ClimateNetwork`.

        .. note::
           Either threshold **OR** link_density have to be given!

        Possible choices for ``node_weight_type``:
          - None (constant unit weights)
          - "surface" (cos lat)
          - "irrigation" (cos**2 lat)

        :type grid: :class:`.GeoGrid`
        :arg  grid: The GeoGrid object describing the network's spatial
            embedding.
        :type similarity_measure: 2D array [index, index]
        :arg similarity_measure: The similarity measure for all pairs of nodes.
        :arg float threshold: The threshold of similarity measure, above which
            two nodes are linked in the network.
        :arg float link_density: The networks's desired link density.
        :arg bool non_local: Determines, whether links between spatially close
            nodes should be suppressed.
        :arg bool directed: Determines, whether the network is treated as
            directed.
        :arg str node_weight_type: The type of geographical node weight to be
            used.
        :arg int silence_level: The inverse level of verbosity of the object.
        """
        #  Initialize
        assert isinstance(grid, GeoGrid)
        self.grid: GeoGrid = grid
        self.directed = directed
        self.silence_level = silence_level

        # mutation count
        if not hasattr(self, "_mut_clim"):
            self._mut_clim: int = 0
        else:
            self._mut_clim += 1

        #  FIXME: Is taking the absolute value by default OK?
        self._similarity_measure = np.abs(similarity_measure.astype("float32"))
        self._non_local = non_local
        self.N = grid.N
        self.node_weight_type = node_weight_type

        #  Sets the threshold and generates the network by thresholding and
        #  calling the "constructor" of parent class GeoNetwork.
        if threshold is not None:
            self.set_threshold(threshold)
        elif link_density is not None:
            self.set_link_density(link_density)
        else:
            print("Either threshold or link_density have to be prescribed "
                  "for network construction!")
        GeoNetwork.__init__(self, adjacency=self.adjacency, grid=self.grid,
                            directed=self.directed,
                            node_weight_type=self.node_weight_type,
                            silence_level=self.silence_level)

    def __cache_state__(self) -> Tuple[Hashable, ...]:
        return GeoNetwork.__cache_state__(self) + (self._mut_clim,)

    def __rec_cache_state__(self) -> Tuple[object, ...]:
        return (self.grid,)

    def __str__(self):
        """
        Return a string representation of the ClimateNetwork object.

        **Example:**

        >>> print(ClimateNetwork.SmallTestNetwork())
        ClimateNetwork:
        GeoNetwork:
        Network: undirected, 6 nodes, 7 links, link density 0.467.
        Geographical boundaries:
                 time     lat     lon
           min    0.0    0.00    2.50
           max    9.0   25.00   15.00
        Threshold: 0.5
        Local connections filtered out: False
        """
        return (f'ClimateNetwork:\n{GeoNetwork.__str__(self)}\n' +
                f'Threshold: {self.threshold()}\n' +
                f'Local connections filtered out: {self.non_local()}')

    def _regenerate_network(self):
        """
        Regenerate the current climate network according to a new similarity
        measure.
        """
        ClimateNetwork.__init__(self, grid=self.data.grid,
                                similarity_measure=self._similarity_measure,
                                threshold=self._threshold,
                                link_density=self.link_density,
                                non_local=self._non_local,
                                directed=self.directed,
                                node_weight_type=self.node_weight_type,
                                silence_level=self.silence_level)

    #
    #  Load and save ClimateNetwork object
    #

    def save(self, filename, fileformat=None, *args, **kwds):
        """
        Save the ClimateNetwork object to files.

        Unified writing function for graphs. Relies on and partially extends
        the corresponding igraph function. Refer to igraph documentation for
        further details on the various writer methods for different formats.

        This method tries to identify the format of the graph given in
        the first parameter (based on extension) and calls the corresponding
        writer method.

        Existing node and link attributes/weights are also stored depending
        on the chosen file format. E.g., the formats GraphML and gzipped
        GraphML are able to store both node and link weights.

        .. note::
           The similarity measure matrix and grid are not stored if
           the corresponding filenames are None.

        The remaining arguments are passed to the writer method without
        any changes.

        :arg tuple/list filename: Tuple or list of three strings, namely
            the paths to the files where the Network object, the
            GeoGrid object and the similarity measure matrix are to be stored.
        :arg str fileformat: the format of the file (if one wants to override
            the format determined from the filename extension, or the filename
            itself is a stream). ``None`` means auto-detection.  Possible
            values are: ``"ncol"`` (NCOL format), ``"lgl"`` (LGL format),
            ``"graphml"``, ``"graphmlz"`` (GraphML and gzipped GraphML format),
            ``"gml"`` (GML format), ``"dot"``, ``"graphviz"`` (DOT format, used
            by GraphViz), ``"net"``, ``"pajek"`` (Pajek format), ``"dimacs"``
            (DIMACS format), ``"edgelist"``, ``"edges"`` or ``"edge"`` (edge
            list), ``"adjacency"`` (adjacency matrix), ``"pickle"`` (Python
            pickled format), ``"svg"`` (Scalable Vector Graphics).
        :arg str filename_similarity_measure:  The name of the file where the
            similarity measure matrix is to be stored.
        """
        try:
            (filename_network, filename_grid,
             filename_similarity_measure) = filename
        except ValueError as e:
            raise ValueError("'filename' must be a tuple or list of three "
                             "items: filename_network, filename_grid, "
                             "filename_similarity_measure") from e

        #  Store GeoNetwork
        GeoNetwork.save(self, filename=(filename_network, filename_grid),
                        fileformat=fileformat,
                        *args, **kwds)

        #  Store similarity measure
        if filename_similarity_measure is not None:
            similarity_measure = self.similarity_measure()
            similarity_measure.dump(filename_similarity_measure)

    @staticmethod
    def Load(filename, fileformat=None, silence_level=0, *args, **kwds):
        """
        Return a ClimateNetwork object stored in files.

        Unified reading function for graphs. Relies on and partially extends
        the corresponding igraph function. Refer to igraph documentation for
        further details on the various reader methods for different formats.

        This method tries to identify the format of the graph given in
        the first parameter and calls the corresponding reader method.

        Existing node and link attributes/weights are also restored depending
        on the chosen file format. E.g., the formats GraphML and gzipped
        GraphML are able to store both node and link weights.

        The remaining arguments are passed to the reader method without
        any changes.

        :arg tuple/list filename: Tuple or list of three strings, namely
            the paths to the files containing the Network object, the
            GeoGrid object and the similarity measure matrix.
            (filename_network, filename_grid, filename_similarity_measure)
        :arg str fileformat: the format of the file (if known in advance)
            ``None`` means auto-detection. Possible values are: ``"ncol"``
            (NCOL format), ``"lgl"`` (LGL format), ``"graphml"``,
            ``"graphmlz"`` (GraphML and gzipped GraphML format), ``"gml"`` (GML
            format), ``"net"``, ``"pajek"`` (Pajek format), ``"dimacs"``
            (DIMACS format), ``"edgelist"``, ``"edges"`` or ``"edge"`` (edge
            list), ``"adjacency"`` (adjacency matrix), ``"pickle"`` (Python
            pickled format).
        :return: :class:`ClimateNetwork` instance.
        """
        try:
            (filename_network, filename_grid,
             filename_similarity_measure) = filename
        except ValueError as e:
            raise ValueError("'filename' must be a tuple or list of three "
                             "items: filename_network, filename_grid, "
                             "filename_similarity_measure") from e

        #  Load GeoGrid object
        grid = GeoGrid.Load(filename_grid)

        #  Load similarity measure
        similarity_measure = np.load(filename_similarity_measure)

        #  Load to igraph Graph object
        graph = igraph.Graph.Read(f=filename_network, format=fileformat,
                                  *args, **kwds)

        #  Extract adjacency matrix
        A = np.array(graph.get_adjacency(type=2).data)

        #  Extract node weights
        if "node_weight_nsi" in graph.vs.attribute_names():
            node_weights = np.array(
                graph.vs.get_attribute_values("node_weight_nsi"))
        else:
            node_weights = None

        #  Create ClimateNetwork instance
        net = ClimateNetwork(grid=grid, similarity_measure=similarity_measure,
                             directed=graph.is_directed(),
                             silence_level=silence_level)
        net.adjacency = A
        net.node_weights = node_weights

        #  Overwrite igraph Graph object in Network instance to restore link
        #  attributes/weights
        net.graph = graph
        #  invalidate cache
        net._mut_la += 1
        return net

    #
    #  Methods for testing purposes
    #

    @staticmethod
    def SmallTestNetwork():
        """
        Return a 6-node undirected test climate network from a similarity
        matrix.

        The network looks like this::

                3 - 1
                |   | \\
            5 - 0 - 4 - 2

        **Example:**

        >>> r(ClimateNetwork.SmallTestNetwork().adjacency)
        array([[0, 0, 0, 1, 1, 1], [0, 0, 1, 1, 1, 0], [0, 1, 0, 0, 1, 0],
               [1, 1, 0, 0, 0, 0], [1, 1, 1, 0, 0, 0], [1, 0, 0, 0, 0, 0]])

        :rtype: :class:`.Network` instance
        """
        return ClimateNetwork(grid=GeoGrid.SmallTestGrid(),
                              similarity_measure=np.array(
                                  [[1.0, 0.1, 0.2, 0.6, 0.7, 0.55],
                                   [0.1, 1.0, 0.55, 0.9, 1.0, 0.3],
                                   [0.2, 0.55, 1.0, 0.2, 0.8, 0.1],
                                   [0.6, 0.9, 0.1, 1.0, 0.1, 0.3],
                                   [0.7, 1.0, 0.8, 0.1, 1.0, 0.4],
                                   [0.55, 0.3, 0.1, 0.3, 0.4, 1.0]]),
                              threshold=0.5,
                              directed=False,
                              silence_level=2)

    #
    #  Methods related to fixing link density
    #

    def link_density_function(self, n_bins):
        """
        Return the network's link density as a function of the threshold.

        **Example:**

        >>> r(ClimateNetwork.SmallTestNetwork().\
                link_density_function(3)[0])
        array([ 0. , 0.3889, 0.6667])
        >>> r(ClimateNetwork.SmallTestNetwork().\
                link_density_function(3)[1])
        array([ 0.1, 0.4, 0.7, 1. ])

        :type n_bins: number (int)
        :arg n_bins: The number of bins.

        :rtype: tuple of two 1D Numpy arrays [bin]
        :return: the network's link density in dependence on threshold.
        """
        if self.silence_level <= 1:
            print("Calculate the link density function...")

        #  Get the histogram of the correlation measure
        (hist, threshold) = np.histogram(self.similarity_measure(),
                                         bins=n_bins)

        #  Normalize histogram
        hist = hist.astype("float64")
        hist /= hist.sum()

        link_density_function = np.empty(n_bins)

        #  Calculate the link density function
        for i in range(n_bins):
            link_density_function[i] = hist[:i].sum()

        return (link_density_function, threshold)

    def threshold_from_link_density(self, link_density):
        """
        Return the threshold for network construction given link density.

        **Example:**

        >>> r(ClimateNetwork.SmallTestNetwork().\
                threshold_from_link_density(link_density=0.5))
        0.4

        :type link_density: number (float)
        :arg link_density: The networks's desired link density.

        :rtype: number (float)
        :return: The threshold of similarity measure, above which
                 two nodes are linked in the network.
        """
        #  Flatten and sort correlation measure matrix
        flat_corr = self.similarity_measure().copy()
        flat_corr = flat_corr.flatten()
        flat_corr.sort()

        #  Get threshold, exclude the entries on the main diagonal here,
        #  since they will not be included in the network anyways!
        threshold = flat_corr[int((1-link_density) * (len(flat_corr)-self.N))]

        #  Clean up
        del flat_corr

        return threshold

    #
    #  Generate adjacency matrix from correlation measure
    #

    def _calculate_threshold_adjacency(self, similarity_measure, threshold):
        """
        Extract the network's adjacency matrix by thresholding.

        The resulting network is a simple graph, i.e., self-loops and
        multiple links are not allowed.

        **Example** (Threshold zero should yield a fully connected network
        given the test similarity matrix):

        >>> net = ClimateNetwork.SmallTestNetwork()
        >>> net._calculate_threshold_adjacency(
        ...     similarity_measure=net.similarity_measure(), threshold=0.0)
        array([[0, 1, 1, 1, 1, 1], [1, 0, 1, 1, 1, 1],
               [1, 1, 0, 1, 1, 1], [1, 1, 1, 0, 1, 1],
               [1, 1, 1, 1, 0, 1], [1, 1, 1, 1, 1, 0]], dtype=int8)

        :type similarity_measure: 2D Numpy array [index, index]
        :arg  similarity_measure: The similarity measure for all pairs of
                                  nodes.

        :type threshold: number (float)
        :arg  threshold: The threshold of similarity measure, above which
                         two nodes are linked in the network.

        :rtype:  2D Numpy array (int8) [index, index]
        :return: the network's adjacency matrix.
        """
        if self.silence_level <= 1:
            print("Extracting network adjacency matrix by thresholding...")

        N = similarity_measure.shape[0]
        A = np.zeros((N, N), dtype="int8")
        A[similarity_measure > threshold] = 1

        #  Set the diagonal of the adjacency matrix to zero -> no self loops
        #  allowed.
        A.flat[::N+1] = 0

        return A

    def _calculate_non_local_adjacency(self, similarity_measure, threshold,
                                       a=20, d_min=0.05):
        """
        Return the adjacency matrix with suppressed spatially local links.

        Physically trivial links between geographically close nodes
        are removed.

        For large a, :math:`d_min` corresponds to the minimum distance for
        which links are allowed to exist.

        **Example:**

        >>> net = ClimateNetwork.SmallTestNetwork()
        >>> net._calculate_non_local_adjacency(
        ...     similarity_measure=net.similarity_measure(),
        ...     threshold=0.5, a=30, d_min=0.20)
        array([[0, 0, 0, 1, 1, 1], [0, 0, 0, 0, 1, 0],
               [0, 0, 0, 0, 0, 0], [1, 0, 0, 0, 0, 0],
               [1, 1, 0, 0, 0, 0], [1, 0, 0, 0, 0, 0]], dtype=int8)

        :type similarity_measure: 2D Numpy array [index, index]
        :arg  similarity_measure: The similarity measure for all pairs of
                                  nodes.

        :type threshold: number (float)
        :arg  threshold: The threshold of similarity measure, above which
                        two nodes are linked in the network.

        :type a: number (float)
        :arg  a: The steepness parameter of the distance weighting function
                 in the transition region from not including any links
                 (weight=0) to including all links (weight=1).

        :type d_min: number (float)
        :arg  d_min: The parameter controlling the minimum distance, above
                     which links can be included in the network
                     (unit radians).

        :rtype:  2D Numpy array (int8) [index, index]
        :return: the network's adjacency matrix.
        """
        if self.silence_level <= 1:
            print("Extracting network adjacency matrix removing local "
                  "connections...")

        weighted_similarity = similarity_measure * \
            (0.5 * (np.tanh(a * (self.grid.angular_distance() - d_min)) + 1))
        # The above line is a function that provides a smooth
        # transition of distance weight, centered around distance d_min.
        # Other sigmoidal type functions could be used as well.

        return self._calculate_threshold_adjacency(weighted_similarity,
                                                   threshold)

    def similarity_measure(self):
        """
        Return the similarity measure used for network construction.

        **Example:**

        >>> r(ClimateNetwork.SmallTestNetwork().similarity_measure()[0,:])
        array([ 1. , 0.1 , 0.2 , 0.6 , 0.7 , 0.55])

        :rtype: 2D Numpy array [index, index]
        :return: The similarity measure for all pairs of nodes.
        """
        try:
            return self._similarity_measure
        except AttributeError as e:
            raise AttributeError("Similarity matrix was deleted "
                                 "earlier and cannot be retrieved.") from e

    def non_local(self):
        """
        Indicate if links between spatially close nodes were suppressed.

        **Example:**

        >>> ClimateNetwork.SmallTestNetwork().non_local()
        False

        :return bool: Determines, whether links between spatially close nodes
            should be suppressed.
        """
        return self._non_local

    def set_non_local(self, non_local):
        """
        Toggle suppression of links between spatially close nodes.

        **Example:**

        >>> net = ClimateNetwork.SmallTestNetwork()
        >>> net.set_non_local(non_local=True)
        >>> r(net.adjacency)
        array([[0, 0, 0, 1, 1, 1], [0, 0, 0, 1, 1, 0], [0, 0, 0, 0, 1, 0],
               [1, 1, 0, 0, 0, 0], [1, 1, 1, 0, 0, 0], [1, 0, 0, 0, 0, 0]])

        :arg bool non_local: Determines, whether links between spatially close
            nodes should be suppressed.
        """
        #  Only change the network if there is a real change in non_local
        if self.non_local() != non_local:
            self._non_local = non_local
            #  Regenerate the climate network using the new setting
            self.set_threshold(self.threshold())

    def threshold(self):
        """
        Return the threshold used to generate the current climate network.

        **Example:**

        >>> ClimateNetwork.SmallTestNetwork().threshold()
        0.5

        :rtype: number (float)
        :return: the threshold used to generate the current climate network.
        """
        return self._threshold

    def set_threshold(self, threshold):
        """
        Generate climate network by thresholding similarity matrix.

        **Example** (Number of links decreases as threshold increases):

        >>> net = ClimateNetwork.SmallTestNetwork()
        >>> net.n_links
        7
        >>> net.set_threshold(threshold=0.7)
        >>> net.n_links
        3

        :type threshold: number (float)
        :arg threshold: the threshold used to generate the current climate
                          network.
        """
        #  Set class variable _threshold
        self._threshold = threshold

        similarity = self.similarity_measure()

        if self.non_local():
            A = self._calculate_non_local_adjacency(similarity, threshold)
        else:
            A = self._calculate_threshold_adjacency(similarity, threshold)

        #  Call constructor of parent class GeoNetwork
        GeoNetwork.__init__(self, adjacency=A, grid=self.grid,
                            directed=self.directed,
                            node_weight_type=self.node_weight_type,
                            silence_level=self.silence_level)

    def set_link_density(self, link_density):
        """
        Generate climate network by thresholding with prescribed link density.

        .. note::
           The desired link density can only be achieved approximately in most
           cases.

        **Example:**

        >>> net = ClimateNetwork.SmallTestNetwork()
        >>> r(net.link_density)
        0.4667
        >>> net.set_link_density(link_density=0.7)
        >>> r(net.link_density)
        0.6667

        :type link_density: number (float)
        :arg link_density: The networks's desired link density.
        """
        threshold = self.threshold_from_link_density(link_density)
        self.set_threshold(threshold)

    @Cached.method()
    def correlation_distance(self):
        """
        Return correlation weighted distances between nodes.

        Defined as the elementwise product of the correlation measure and
        angular great circle distance matrices.

        This is a useful measure of the relative importance of links,
        since links with high geographical distance and high correlation
        (teleconnections) get the highest weight. Trivial correlations with
        small geographical distance and high correlation get a lower weight.

        Correlation distance appears to be the simplest functional form of
        combining geographical distance and correlation measure that yields
        meaningful results.

        **Example:**

        >>> r(ClimateNetwork.SmallTestNetwork().correlation_distance(), 2)
        array([[ 0.  , 0.01, 0.04, 0.18, 0.27, 0.27],
               [ 0.01, 0.  , 0.05, 0.18, 0.29, 0.12],
               [ 0.04, 0.05, 0.  , 0.02, 0.16, 0.03],
               [ 0.18, 0.18, 0.01, 0.  , 0.01, 0.06],
               [ 0.27, 0.29, 0.16, 0.01, 0.  , 0.04],
               [ 0.27, 0.12, 0.03, 0.06, 0.04, 0.  ]])

        :rtype: 2D matrix [index, index]
        :return: the correlation distance matrix.
        """
        return self.similarity_measure() * self.grid.angular_distance()

    @Cached.method()
    def inv_correlation_distance(self):
        """
        Return correlation weighted distances between nodes.

        :rtype: 2D matrix [index, index]
        """
        m = self.correlation_distance()
        np.fill_diagonal(m, np.inf)
        self.set_link_attribute('inv_correlation_distance', 1 / m)
        return 1 / m

    #
    #  Link weighted network measures
    #

    def correlation_distance_weighted_closeness(self):
        """
        Return correlation distance weighted closeness.

        Calculates the sequence of closeness centralities link-weighted by the
        inverse of correlation distance between nodes. For closeness
        centrality calculation, the inverse of correlation distance is used,
        because high values of this measure should correspond to short
        distances in the graph and vice versa when weighted shortest paths are
        calculated.

        **Example:**

        >>> r(ClimateNetwork.SmallTestNetwork().\
                correlation_distance_weighted_closeness())
        array([ 0.1646, 0.1351, 0.0894, 0.1096, 0.1659, 0.1102])

        :rtype: 1D Numpy array [index]
        :return: the correlation distance weighted closeness sequence.
        """
        self.inv_correlation_distance()
        return self.closeness('inv_correlation_distance')

    def local_correlation_distance_weighted_vulnerability(self):
        """
        Return local correlation distance weighted vulnerability.

        Calculates the sequence of vulnerabilities link-weighted by the
        inverse of correlation distance between nodes. For vulnerability
        calculation, the inverse of correlation distance is used, because
        high values of this measure should correspond to short distances in
        the graph and vice versa when weighted shortest paths are calculated.

        **Example:**

        >>> r(ClimateNetwork.SmallTestNetwork().\
                local_correlation_distance_weighted_vulnerability())
        array([ 0.4037, 0.035 , -0.1731, -0.081 , 0.3121, -0.0533])

        :rtype: 1D Numpy array
        :return: the local correlation distance weighted vulnerability
                 sequence.
        """
        self.inv_correlation_distance()
        return self.local_vulnerability('inv_correlation_distance')

    def _weighted_metric(self, attr: str, calc: Callable, metric: str):
        if not self.find_link_attribute(attr):
            self.set_link_attribute(attr, calc())
        return getattr(self, metric)(attr)



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Provides class for event series analysis, namely event synchronization (ES) and
event coincidence analysis (ECA). In addition, a method for the generation of
binary event series from continuous time series data is included.
When instantiating a class, data must either be passed as an event
matrix (for details see below) or as a continuous time series. Using the class,
an ES or ECA matrix can be calculated to generate a climate network using the
EventSeriesClimateNetwork class. Both ES and ECA may be called without
instantiating an object of the class.
Significance levels are provided using analytic calculations using Poisson
point processes as a null model (for ECA only) or a Monte Carlo approach.
"""

from typing import Tuple
from collections.abc import Hashable

import warnings

import numpy as np
from scipy import stats

from ..core.cache import Cached


class EventSeries(Cached):

    def __init__(self, data, timestamps=None, taumax=np.inf, lag=0.0,
                 threshold_method=None, threshold_values=None,
                 threshold_types=None):
        """
        Initialize an instance of EventSeries. Input data must be a 2D numpy
        array with time as the first axis and variables as the second axis.
        Event data is stored as an eventmatrix.

        Format of eventmatrix:
        An eventmatrix is a 2D numpy array with the first dimension covering
        the timesteps and the second dimensions covering the variables. Each
        variable at a specific timestep is either '1' if an event occured or
        '0' if it did not, e.g. for 3 variables with 10 timesteps the
        eventmatrix could look like

            array([[0, 1, 0],
                   [1, 0, 1],
                   [0, 0, 0],
                   [1, 0, 1],
                   [0, 1, 0],
                   [0, 0, 0],
                   [1, 0, 0],
                   [0, 0, 1],
                   [0, 1, 0],
                   [0, 0, 0]])

        If input data is not provided as an eventmatrix, the constructor tries
        to generate one using the make_event_matrix method. Default keyword
        arguments are used in this case.

        :type data: 2D Numpy array [time, variables]
        :arg data: Event series array or array of non-binary variable values
        :type timestamps: 1D Numpy array
        :arg timestamps: Time points of events of data. If not provided,
                        integer values are used
        :type taumax: float
        :arg taumax: maximum time difference between two events to be
                    considered synchronous. Caution: For ES, the default is
                    np.inf because of the intrinsic dynamic coincidence
                    interval in ES. For ECA, taumax is a parameter of the
                    method that needs to be defined!
        :type lag: float
        :arg lag: extra time lag between the event series
        :type threshold_method: str 'quantile' or 'value' or 1D numpy array or
                                str 'quantile' or 'value'
        :arg threshold_method: specifies the method for generating a binary
                               event matrix from an array of continuous time
                               series. Default: None
        :type threshold_values: 1D Numpy array or float
        :arg threshold_values: quantile or real number determining threshold
                               for each variable. Default: None.
        :type threshold_types: str 'above' or 'below' or 1D list of strings
                               'above' or 'below'
        :arg threshold_types: Determines for each variable if event is below
                              or above threshold
        """

        if threshold_method is None:
            # Check if data contains only binary values
            if len(np.unique(data)) != 2 or not (
                    np.unique(data) == np.array([0, 1])).all():
                raise IOError("Event matrix not in correct format")

            # Save event matrix
            self.__T = data.shape[0]
            self.__N = data.shape[1]
            self.__eventmatrix = data

        else:
            # If data is not in eventmatrix format, use method
            # make_event_matrix to transform continuous time series to a binary
            # time series
            # Allow for wrong axis, i.e. first axis variables and second axis
            # time if time series have the same length
            if isinstance(data, np.ndarray):
                if data.shape[1] > data.shape[0]:
                    data = np.swapaxes(data, 0, 1)

                self.__eventmatrix = \
                    self.make_event_matrix(data,
                                           threshold_method=threshold_method,
                                           threshold_values=threshold_values,
                                           threshold_types=threshold_types)

                self.__T = self.__eventmatrix.shape[0]
                self.__N = self.__eventmatrix.shape[1]

            else:
                raise IOError('Input data is not in event matrix format!')

        # If no timestamps are given, use integer array indices as timestamps
        if timestamps is not None:
            if timestamps.shape[0] != self.__T:
                raise IOError("Timestamps array has not the same length as"
                              " event matrix!")
            self.__timestamps = timestamps
        else:
            self.__timestamps = np.linspace(0.0, self.__T - 1, self.__T)

        self.__taumax = float(taumax)
        self.__lag = float(lag)

        # save number of events
        NrOfEvs = np.array(np.sum(self.__eventmatrix, axis=0), dtype=int)
        self.__nrofevents = NrOfEvs

        # Dictionary of symmetrization functions for later use
        self.symmetrization_options = {
            'directed': EventSeries._symmetrization_directed,
            'symmetric': EventSeries._symmetrization_symmetric,
            'antisym': EventSeries._symmetrization_antisym,
            'mean': EventSeries._symmetrization_mean,
            'max': EventSeries._symmetrization_max,
            'min': EventSeries._symmetrization_min
        }

    def __cache_state__(self) -> Tuple[Hashable, ...]:
        # The following attributes are assumed immutable:
        #   (__eventmatrix, __timestamps, __taumax, __lag)
        return ()

    def __str__(self):
        """
        Return a string representation of the EventSeries object.
        """
        return (f"EventSeries: {self.__N} variables, "
                f"{self.__T} timesteps, taumax: {self.__taumax:.1f}, "
                f"lag: {self.__lag:.1f}")

    def get_event_matrix(self):
        return self.__eventmatrix

    @staticmethod
    def _symmetrization_directed(matrix):
        """
        Helper function for symmetrization options

        :type matrix: 2D numpy array
        :arg matrix: pairwise ECA/ES scores of data
        :rtype: 2D numpy array
        :return: original matrix
        """
        return matrix

    @staticmethod
    def _symmetrization_symmetric(matrix):
        """
        Helper function for symmetrization options

        :type matrix: 2D numpy array
        :arg matrix: pairwise ECA/ES scores of data
        :rtype: 2D numpy array
        :return: symmetrized matrix
        """
        return matrix + matrix.T

    @staticmethod
    def _symmetrization_antisym(matrix):
        """
        Helper function for symmetrization options

        :type matrix: 2D numpy array
        :arg matrix: pairwise ECA/ES scores of data
        :rtype: 2D numpy array
        :return: antisymmetrized matrix
        """
        return matrix - matrix.T

    @staticmethod
    def _symmetrization_mean(matrix):
        """
        Helper function for symmetrization options

        :type matrix: 2D numpy array
        :arg matrix: pairwise ECA/ES scores of data
        :rtype: 2D numpy array
        :return: symmetrized matrix using element-wise mean of matrix and
                 transposed matrix
        """
        return np.mean([matrix, matrix.T], axis=0)

    @staticmethod
    def _symmetrization_max(matrix):
        """
        Helper function for symmetrization options

        :type matrix: 2D numpy array
        :arg matrix: pairwise ECA/ES scores of data
        :rtype: 2D numpy array
        :return: symmetrized matrix using element-wise maximum of matrix and
                 transposed matrix
        """
        return np.maximum(matrix, matrix.T)

    @staticmethod
    def _symmetrization_min(matrix):
        """
        Helper function for symmetrization options

        :type matrix: 2D numpy array
        :arg matrix: pairwise ECA/ES scores of data
        :rtype: 2D numpy array
        :return: symmetrized matrix using element-wise minimum of matrix and
                 transposed matrix
        """
        return np.minimum(matrix, matrix.T)

    @staticmethod
    def make_event_matrix(data, threshold_method='quantile',
                          threshold_values=None, threshold_types=None):
        """
        Create a binary event matrix from continuous time series data. Data
        format is eventmatrix, i.e. a 2D numpy array with first dimension
        covering time and second dimension covering the values of the
        variables.

        :type data: 2D numpy array
        :arg data: Continuous input data
        :type threshold_method: str 'quantile' or 'value' or 1D numpy array of
                                strings 'quantile' or 'value'
        :arg threshold_method: specifies the method for generating a binary
                               event matrix from an array of continuous time
                               series. Default: 'quantile'
        :type threshold_values: 1D Numpy array or float
        :arg threshold_values: quantile or real number determining threshold
                               for each variable. Default: None.
        :type threshold_types: str 'above' or 'below' or 1D list of strings
                               'above' or 'below'
        :arg threshold_types: Determines for each variable if event is below or
                              above threshold
        :rtype: 2D numpy array
        :return: eventmatrix
        """

        # Check correct format of event matrix
        if not np.all([len(i) == len(data[0]) for i in data]):
            warnings.warn("Data does not contain equal number of events")

        data_axswap = np.swapaxes(data, 0, 1)
        thresholds = np.zeros(data.shape[1])

        # Check if inserted keyword arguments are correct and create parameter
        # arrays in case only single keywords are used for data with more than
        # one variable
        threshold_method = np.array(threshold_method)
        if threshold_method.shape == (data.shape[1],):
            if not np.all([i in ['quantile', 'value'] for i in
                           threshold_method]):
                raise IOError("'threshold_method' must be either 'quantile' or"
                              " 'value' or a 1D array-like object with"
                              " entries 'quantile' or 'value' for each"
                              " variable!")
        elif not threshold_method.shape:
            if threshold_method in ['quantile', 'value']:
                threshold_method = np.array([threshold_method] * data.shape[1])
            else:
                raise IOError("'threshold_method' must be either 'quantile' or"
                              " 'value' or a 1D array-like object with entries"
                              " 'quantile' or 'value' for each variable!")
        else:
            raise IOError("'threshold_method' must be either 'quantile' or "
                          "'value' or a 1D array-like object with entries "
                          "'quantile' or 'value' for each variable!")

        if threshold_values is not None:
            threshold_values = np.array(threshold_values)
            if threshold_values.shape == (data.shape[1],):
                if not np.all([isinstance(i, (float, int))
                               for i in threshold_values]):
                    raise IOError("'threshold_values' must be either float/int"
                                  " or 1D array-like object of float/int for "
                                  " each variable!")
            elif not threshold_values.shape:
                if isinstance(threshold_values.item(), (int, float)):
                    threshold_values = \
                        np.array([threshold_values] * data.shape[1])
                else:
                    raise IOError("'threshold_values' must be either float/int"
                                  " or 1D array-like object of float/int for "
                                  "each variable!")
            else:
                raise IOError("'threshold_values' must be either float/int or "
                              "1D array-like object of float/int for each "
                              "variable!")
        else:
            threshold_values = np.array([None] * data.shape[1])
            warnings.warn("No 'threshold_values' given. Median is used by "
                          "default!")

        if threshold_types is not None:
            threshold_types = np.array(threshold_types)
            if threshold_types.shape == (data.shape[1],):
                if not np.all([i in ['above', 'below']
                               for i in threshold_types]):
                    raise IOError("'threshold_types' must be either 'above' or"
                                  " 'below' or a 1D array-like object with "
                                  "entries 'above' or 'below' for each "
                                  "variable!")
            elif not threshold_types.shape:
                if threshold_types in ['above', 'below']:
                    threshold_types = \
                        np.array([threshold_types] * data.shape[1])
                else:
                    raise IOError("'threshold_types' must be either 'above' or"
                                  " 'below' or a 1D array-like object with "
                                  "entries 'above' or 'below' for each "
                                  "variable!")
            else:
                raise IOError("'threshold_types' must be either 'above' or "
                              "'below' or a 1D array-like object with entries "
                              "'above' or 'below' for each variable!")
        else:
            threshold_types = np.array([None] * data.shape[1])
            warnings.warn("No 'threshold_types' given. If 'threshold_values' "
                          ">= median, 'above' is used by default!")

        # Go through threshold_method, threshold_value and threshold_type
        # for each variable and check if input parameters are valid
        # In case of missing input parameters, try to set default values
        for i in range(data.shape[1]):

            if threshold_method[i] == 'quantile':

                # Check if threshold quantile is between zero and one
                if threshold_values[i] is not None:
                    if threshold_values[i] > 1.0 or threshold_values[i] < 0.0:
                        raise ValueError("Threshold_value for threshold_method"
                                         " 'quantile' must lie between 0.0 and"
                                         " 1.0!")

                # If threshold values are not given, use the median
                else:
                    threshold_values[i] = 0.5

                # Compute threshold value according to quantile
                thresholds[i] = \
                    np.quantile(data_axswap[i], threshold_values[i])

                # If no threshold_types is given, check if threshold value is
                # larger or equal median, then 'above'
                if threshold_types[i] is None:
                    if threshold_values[i] >= 0.5:
                        threshold_types[i] = 'above'
                    else:
                        threshold_types[i] = 'below'

            if threshold_method[i] == 'value':

                if threshold_values[i] is None:
                    thresholds[i] = np.median(data_axswap[i])
                else:
                    # Check if given threshold values lie within data range
                    if np.max(data_axswap[i]) < threshold_values[i] or \
                            np.min(data_axswap[i]) > threshold_values[i]:
                        raise IOError("Threshold_value for threshold_method "
                                      "'value' must lie within variable "
                                      "range!")
                    thresholds[i] = threshold_values[i]

                if threshold_types[i] is None:
                    if thresholds[i] >= np.median(data_axswap[i]):
                        threshold_types[i] = 'above'
                    else:
                        threshold_types[i] = 'below'

        # Other methods for thresholding can be easily added here

        eventmatrix = np.zeros((data.shape[0], data.shape[1])) * (-1)
        # Iterate through all variables of the data and create event matrix
        # according to specified methods
        for t in range(data.shape[0]):
            for i in range(data.shape[1]):
                if threshold_types[i] == 'above':
                    if data[t][i] > thresholds[i]:
                        eventmatrix[t][i] = 1
                    else:
                        eventmatrix[t][i] = 0
                elif threshold_types[i] == 'below':
                    if data[t][i] < thresholds[i]:
                        eventmatrix[t][i] = 1
                    else:
                        eventmatrix[t][i] = 0

        return eventmatrix

    @staticmethod
    def event_synchronization(eventseriesx, eventseriesy, ts1=None, ts2=None,
                              taumax=np.inf, lag=0.0):
        """
        Calculates the directed event synchronization from two event series X
        and Y using the algorithm described in [Quiroga2002]_,
        [Odenweller2020]_

        :type eventseriesx: 1D Numpy array
        :arg eventseriesx: Event series containing '0's and '1's
        :type eventseriesy: 1D Numpy array
        :arg eventseriesy: Event series containing '0's and '1's
        :type ts1: 1D Numpy array
        :arg ts1: Event time array containing time points when events of event
                  series 1 occur, not obligatory
        :type ts2: 1D Numpy array
        :arg ts2: Event time array containing time points when events of event
                  series 2 occur, not obligatory
        :type taumax: float
        :arg taumax: maximum distance of two events to be counted as
                     synchronous
        :type lag: float
        :arg lag: delay between the two event series, the second event series
                  is shifted by the value of lag
        :rtype: list
        :return: [Event synchronization XY, Event synchronization YX]
        """

        # Get time indices (type boolean or simple '0's and '1's)
        # Careful here with datatype, int16 allows for maximum time index 32767
        # Get time indices
        if ts1 is None:
            ex = np.array(np.where(eventseriesx), dtype='int16')
        else:
            ex = np.array([ts1[eventseriesx == 1]], dtype='float')
        if ts2 is None:
            ey = np.array(np.where(eventseriesy), dtype='int16')
        else:
            ey = np.array([ts2[eventseriesy == 1]], dtype='float')

        ey = ey + lag

        lx = ex.shape[1]
        ly = ey.shape[1]
        if lx == 0 or ly == 0:  # Division by zero in output
            return np.nan, np.nan
        if lx in [1, 2] or ly in [1, 2]:  # Too few events to calculate
            return 0., 0.

        # Array of distances
        dstxy2 = 2 * (np.repeat(ex[:, 1:-1].T, ly - 2, axis=1)
                      - np.repeat(ey[:, 1:-1], lx - 2, axis=0))
        # Dynamical delay
        diffx = np.diff(ex)
        diffy = np.diff(ey)
        diffxmin = np.minimum(diffx[:, 1:], diffx[:, :-1])
        diffymin = np.minimum(diffy[:, 1:], diffy[:, :-1])
        tau2 = np.minimum(np.repeat(diffxmin.T, ly - 2, axis=1),
                          np.repeat(diffymin, lx - 2, axis=0))
        tau2 = np.minimum(tau2, 2 * taumax)

        # Count equal time events and synchronised events
        eqtime = dstxy2.size - np.count_nonzero(dstxy2)

        # Calculate boolean matrices of coincidences
        Axy = (dstxy2 > 0) * (dstxy2 <= tau2)
        Ayx = (dstxy2 < 0) * (dstxy2 >= -tau2)

        # Loop over coincidences and determine number of double counts
        # by checking at least one event of the pair is also coincided
        # in other direction
        countxydouble = countyxdouble = 0

        for i, j in np.transpose(np.where(Axy)):
            countxydouble += np.any(Ayx[i, :]) or np.any(Ayx[:, j])
        for i, j in np.transpose(np.where(Ayx)):
            countyxdouble += np.any(Axy[i, :]) or np.any(Axy[:, j])

        # Calculate counting quantities and subtract half of double countings
        countxy = np.sum(Axy) + 0.5 * eqtime - 0.5 * countxydouble
        countyx = np.sum(Ayx) + 0.5 * eqtime - 0.5 * countyxdouble

        norm = np.sqrt((lx - 2) * (ly - 2))
        return countxy / norm, countyx / norm

    @staticmethod
    def event_coincidence_analysis(eventseriesx, eventseriesy, taumax,
                                   ts1=None, ts2=None, lag=0.0):
        """
         Event coincidence analysis:
         Returns the precursor and trigger coincidence rates of two event
         series X and Y following the algorithm described in [Odenweller2020]_.

         :type eventseriesx: 1D Numpy array
         :arg eventseriesx: Event series containing '0's and '1's
         :type eventseriesy: 1D Numpy array
         :arg eventseriesy: Event series containing '0's and '1's
         :type ts1: 1D Numpy array
         :arg ts1: Event time array containing time points when events of event
                   series 1 occur, not obligatory
         :type ts2: 1D Numpy array
         :arg ts2: Event time array containing time points when events of event
                   series 2 occur, not obligatory
         :type taumax: float
         :arg taumax: coincidence interval width
         :type lag: int
         :arg lag: lag parameter
         :rtype: list
         :return: [Precursor coincidence rate XY, Trigger coincidence rate XY,
               Precursor coincidence rate YX, Trigger coincidence rate YX]
         """

        # Get time indices
        if ts1 is None:
            e1 = np.where(eventseriesx)[0]
        else:
            e1 = ts1[eventseriesx == 1]
        if ts2 is None:
            e2 = np.where(eventseriesy)[0]
        else:
            e2 = ts2[eventseriesy == 1]

        # Count events that cannot be coincided due to lag and delT
        if not (lag == 0 and taumax == 0):
            n11 = len(e1[e1 <= e1[0] + lag + taumax])  # Start of es1
            n12 = len(e1[e1 >= (e1[-1] - lag - taumax)])  # End of es1
            n21 = len(e2[e2 <= e2[0] + lag + taumax])  # Start of es2
            n22 = len(e2[e2 >= (e2[-1] - lag - taumax)])  # End of es2
        else:
            n11, n12, n21, n22 = 0, 0, 0, 0  # Instantaneous coincidence

        # Number of events
        l1 = len(e1)
        l2 = len(e2)
        # Array of all interevent distances
        dst = (np.array([e1] * l2).T - np.array([e2] * l1))

        # Count coincidences with array slicing
        prec12 = np.count_nonzero(
            np.any(((dst - lag >= 0) * (dst - lag <= taumax))[n11:, :],
                   axis=1))
        trig12 = np.count_nonzero(
            np.any(((dst - lag >= 0) * (dst - lag <= taumax))
                   [:, :dst.shape[1] - n22], axis=0))
        prec21 = np.count_nonzero(np.any(((-dst - lag >= 0)
                                          * (-dst - lag <= taumax))[:, n21:],
                                         axis=0))
        trig21 = np.count_nonzero(
            np.any(((-dst - lag >= 0) * (-dst - lag <= taumax))
                   [:dst.shape[0] - n12, :], axis=1))

        # Normalisation and output
        return (np.float32(prec12) / (l1 - n11),
                np.float32(trig12) / (l2 - n22),
                np.float32(prec21) / (l2 - n21),
                np.float32(trig21) / (l1 - n12))

    def _eca_coincidence_rate(self, eventseriesx, eventseriesy,
                              window_type='symmetric', ts1=None, ts2=None):
        """
         Event coincidence analysis:
         Returns the coincidence rates of two event series for both directions

         :type eventseriesx: 1D Numpy array
         :arg eventseriesx: Event series containing '0's and '1's
         :type eventseriesy: 1D Numpy array
         :arg eventseriesy: Event series containing '0's and '1's
         :type ts1: 1D Numpy array
         :arg ts1: Event time array containing time points when events of event
                   series 1 occur, not obligatory
         :type ts2: 1D Numpy array
         :arg ts2: Event time array containing time points when events of event
                   series 2 occur, not obligatory
         :type window_type: str {'retarded', 'advanced', 'symmetric'}
         :arg window_type: Only for ECA. Determines if precursor coincidence
                           rate ('advanced'), trigger coincidence rate
                           ('retarded') or a general coincidence rate with the
                           symmetric interval [-taumax, taumax] are computed
                           ('symmetric'). Default: 'symmetric'
         :rtype: list
         :return: Precursor coincidence rates [XY, YX]
         """
        # Get time indices
        if ts1 is None:
            e1 = np.where(eventseriesx)[0]
        else:
            e1 = ts1[eventseriesx == 1]
        if ts2 is None:
            e2 = np.where(eventseriesy)[0]
        else:
            e2 = ts2[eventseriesy == 1]

        lag = self.__lag
        taumax = self.__taumax

        # Number of events
        l1 = len(e1)
        l2 = len(e2)

        # Array of all interevent distances
        dst = (np.array([e1] * l2).T - np.array([e2] * l1))

        if window_type == 'advanced':
            deltaT1 = 0.0
            deltaT2 = taumax

            # Count events that cannot be coincided due to lag and deltaT
            if not (lag == 0 and taumax == 0):
                n11 = len(e1[e1 <= (e1[0] + lag + deltaT2)])  # Start of es1
                n21 = len(e2[e2 <= (e2[0] + lag + deltaT2)])  # Start of es2
                n12, n22 = 0, 0
            else:
                n11, n12, n21, n22 = 0, 0, 0, 0  # Instantaneous coincidence

            # Count coincidences with array slicing
            coincidence12 = np.count_nonzero(
                np.any(((dst - lag >= deltaT1) * (dst - lag <= deltaT2))
                       [n11:, :], axis=1))
            coincidence21 = np.count_nonzero(
                np.any(((-dst - lag >= deltaT1) * (-dst - lag <= deltaT2))
                       [:, n21:], axis=0))

        elif window_type == 'retarded':
            deltaT1 = 0.0
            deltaT2 = taumax

            # Count events that cannot be coincided due to lag and delT
            if not (lag == 0 and taumax == 0):
                n11 = 0  # Start of es1
                n12 = len(e1[e1 >= (e1[-1] - lag - deltaT2)])  # End of es1
                n21 = 0  # Start of es2
                n22 = len(e2[e2 >= (e2[-1] - lag - deltaT2)])  # End of es2
            else:
                n11, n12, n21, n22 = 0, 0, 0, 0  # Instantaneous coincidence

            # Count coincidences with array slicing
            coincidence12 = np.count_nonzero(
                np.any(((dst - lag >= deltaT1) * (dst - lag <= deltaT2))
                       [:, :dst.shape[1] - n22], axis=0))
            coincidence21 = np.count_nonzero(
                np.any(((-dst - lag >= deltaT1) * (-dst - lag <= deltaT2))
                       [:dst.shape[0] - n12, :], axis=1))

            return ((np.float32(coincidence12) / (l2 - n22),
                     np.float32(coincidence21) / (l1 - n12)))

        elif window_type == 'symmetric':
            deltaT1, deltaT2 = -taumax, taumax

            # Count events that cannot be coincided due to lag and delT
            if not (lag == 0 and taumax == 0):
                n11 = len(e1[e1 <= (e1[0] + lag + deltaT2)])  # Start of es1
                n12 = len(e1[e1 >= (e1[-1] - lag + deltaT1)])  # End of es1
                n21 = len(e2[e2 <= (e2[0] + lag + deltaT2)])  # Start of es2
                n22 = len(e2[e2 >= (e2[-1] - lag + deltaT1)])  # End of es2
            else:
                n11, n12, n21, n22 = 0, 0, 0, 0  # Instantaneous coincidence

            # Count coincidences with array slicing
            coincidence12 = np.count_nonzero(
                np.any(((dst - lag >= deltaT1) * (dst - lag <= deltaT2))
                       [n11:dst.shape[0]-n12, :], axis=1))
            coincidence21 = np.count_nonzero(
                np.any(((-dst - lag >= deltaT1) * (-dst - lag <= deltaT2))
                       [:, n21:dst.shape[1]-n22], axis=0))

        else:
            raise IOError("Parameter 'window_type' must be 'advanced',"
                          " 'retarded' or 'symmetric'!")

        # Normalisation and output
        return (np.float32(coincidence12) / (l1 - n11 - n12),
                np.float32(coincidence21) / (l2 - n21 - n22))

    def event_series_analysis(self, method='ES', symmetrization='directed',
                              window_type='symmetric'):
        """
        Returns the NxN matrix of the chosen event series measure where N is
        the number of variables. The entry [i, j] denotes the event
        synchronization or event coincidence analysis from variable j to
        variable i. According to the 'symmetrization' parameter the event
        series measure matrix is symmetrized or not.

        The event coincidence rate of ECA is calculated according to the
        formula: r(Y|X, DeltaT1, DeltaT2, tau) =
        1/N_X sum_{i=1}^{N_X} Theta[sum{j=1}^{N_Y}
        1_[DeltaT1, DeltaT2] (t_i^X - (t_j^Y + tau))],
        where X is the first input event series, Y the second input event
        series, N_X the number of events in X, DeltaT1 and DeltaT2 the given
        coincidence interval boundaries, tau the lag between X and Y, Theta the
        Heaviside function and 1 the indicator function.

        :type method: str 'ES' or 'ECA'
        :arg method: determines if ES or ECA should be used
        :type symmetrization: str {'directed', 'symmetric', 'antisym',
                              'mean', 'max', 'min'} for ES,
                              str {'directed', 'mean', 'max', 'min'} for ECA
        :arg symmetrization: determines if and which symmetrisation
                             method should be used for the ES/ECA score matrix
        :type window_type: str {'retarded', 'advanced', 'symmetric'}
        :arg window_type: Only for ECA. Determines if precursor coincidence
                          rate ('advanced'), trigger coincidence rate
                          ('retarded') or a general coincidence rate with the
                          symmetric interval [-taumax, taumax] are computed
                          ('symmetric'). Default: 'symmetric'
        :rtype: 2D numpy array
        :return: pairwise event synchronization or pairwise coincidence rates
                symmetrized according to input parameter 'symmetrization'
        """

        if method not in ['ES', 'ECA']:
            raise IOError("'method' parameter must be 'ECA' or 'ES'!")

        directedESMatrix = []

        if method == 'ES':

            if symmetrization not in ['directed', 'symmetric', 'antisym',
                                      'mean', 'max', 'min']:
                raise IOError("'symmetrization' parameter must be 'directed', "
                              "'symmetric', 'antisym', 'mean', 'max' or"
                              "'min' for event synchronization!")

            directedESMatrix = self._ndim_event_synchronization()

        elif method == 'ECA':
            if self.__taumax is np.inf:
                raise ValueError("'delta' must be a finite time window to"
                                 " determine event coincidence!")

            if symmetrization not in ['directed', 'mean', 'max', 'min']:
                raise IOError("'symmetrization' parameter must be 'directed', "
                              "'mean', 'max' or 'min' for event"
                              "coincidence analysis!")

            if window_type not in ['retarded', 'advanced', 'symmetric']:
                raise IOError("'window_type' must be 'retarded',"
                              "'advanced' or 'symmetric'!")

            directedESMatrix = \
                self._ndim_event_coincidence_analysis(window_type=window_type)

        # Use symmetrization functions for symmetrization and return result
        return self.symmetrization_options[symmetrization](directedESMatrix)

    @Cached.method()
    def _ndim_event_synchronization(self):
        """
        Compute NxN event synchronization matrix [i,j] with event
        synchronization from j to i without symmetrization.

        :rtype: NxN numpy array where N is the number of variables of the
                eventmatrix
        :return: event synchronization matrix
        """
        # Get instance variables
        eventmatrix = self.__eventmatrix
        timestamps = self.__timestamps
        lag = self.__lag
        taumax = self.__taumax

        directed = np.zeros((self.__N, self.__N))
        for i in range(0, self.__N):
            for j in range(i + 1, self.__N):
                directed[i, j], directed[j, i] = \
                    self.event_synchronization(eventmatrix[:, i],
                                               eventmatrix[:, j],
                                               ts1=timestamps, ts2=timestamps,
                                               taumax=taumax, lag=lag)
        return directed

    def _ndim_event_coincidence_analysis(self, window_type='symmetric'):
        """
        Computes NxN event coincidence matrix of event coincidence rate

        :type window_type: str {'retarded', 'advanced', 'symmetric'}
        :arg window_type: Only for ECA. Determines if precursor coincidence
                          rate ('advanced'), trigger coincidence rate
                          ('retarded') or a general coincidence rate with the
                          symmetric interval [-taumax, taumax] are computed
                          ('symmetric'). Default: 'symmetric'
        :rtype: NxN numpy array where N is the number of variables of the
                eventmatrix
        :return: event coincidence matrix
        """

        eventmatrix = self.__eventmatrix
        directed = np.zeros((self.__N, self.__N))
        timestamps = self.__timestamps

        for i in range(0, self.__N):
            for j in range(i + 1, self.__N):
                directed[i, j], directed[j, i] = \
                    self._eca_coincidence_rate(eventmatrix[:, i],
                                               eventmatrix[:, j],
                                               window_type=window_type,
                                               ts1=timestamps, ts2=timestamps)

        return directed

    def _empirical_percentiles(self, method=None, n_surr=1000,
                               symmetrization='directed',
                               window_type='symmetric'):
        """
        Compute p-values of event synchronisation (ES) and event coincidence
        analysis (ECA) using a Monte-Carlo approach. Surrogates are obtained by
        shuffling the event series. ES/ECA scores of the surrogate event series
        are computed and p-values are the empirical percentiles of the original
        event series compared to the ES/ECA scores of the surrogates.

        :type method: str 'ES' or 'ECA'
        :arg method: determines if ES or ECA should be used
        :type n_surr: int
        :arg n_surr: number of surrogates for Monte-Carlo method
        :type symmetrization: str {'directed', 'symmetric', 'antisym',
                              'mean', 'max', 'min'} for ES,
                              str {'directed', 'mean', 'max', 'min'} for ECA
        :arg symmetrization: determines if and which symmetrisation
                             method should be used for the ES/ECA score matrix
        :type window_type: str {'retarded', 'advanced', 'symmetric'}
        :arg window_type: Only for ECA. Determines if precursor coincidence
                          rate ('advanced'), trigger coincidence rate
                          ('retarded') or a general coincidence rate with the
                          symmetric interval [-taumax, taumax] are computed
                          ('symmetric'). Default: 'symmetric'
        :rtype: 2D numpy array
        :return: p-values of the ES/ECA scores for all
        """

        # Get instance variables
        lag = self.__lag
        deltaT = self.__taumax

        event_series_result = \
            self.event_series_analysis(method=method,
                                       symmetrization=symmetrization,
                                       window_type=window_type)

        surrogates = np.zeros((n_surr, self.__N, self.__N))
        shuffled_es = self.__eventmatrix.copy()

        # For each surrogate, shuffle each event series and perform ES/ECA
        # analysis
        for n in range(n_surr):
            for i in range(self.__N):
                np.random.shuffle(shuffled_es[:, i])

            if method == 'ES':
                for i in range(0, self.__N):
                    for j in range(i + 1, self.__N):
                        surrogates[n, i, j], surrogates[n, j, i] = \
                            self.event_synchronization(shuffled_es[:, i],
                                                       shuffled_es[:, j],
                                                       taumax=deltaT, lag=lag)

            elif method == 'ECA':
                for i in range(0, self.__N):
                    for j in range(i + 1, self.__N):
                        surrogates[n, i, j], surrogates[n, j, i] = \
                            self._eca_coincidence_rate(shuffled_es[:, i],
                                                       shuffled_es[:, j],
                                                       window_type=window_type)

            # Symmetrize according to symmetry keyword argument
            surrogates[n, :, :] = \
                self.symmetrization_options[
                    symmetrization](surrogates[n, :, :])

        # Calculate significance level via strict empirical percentiles for
        # each event series pair
        empirical_percentiles = np.zeros((self.__N, self.__N))
        for i in range(self.__N):
            for j in range(self.__N):
                empirical_percentiles[i, j] = \
                    stats.percentileofscore(surrogates[:, i, j],
                                            event_series_result[i][j],
                                            kind='strict') / 100

        return empirical_percentiles

    def event_analysis_significance(self, method=None,
                                    surrogate='shuffle', n_surr=1000,
                                    symmetrization='directed',
                                    window_type='symmetric'):
        """
        Returns significance levels (1 - p-values) for event synchronisation
        (ES) and event coincidence analysis (ECA). For ECA, there is an
        analytic option providing significance levels based on independent
        Poisson processes. The 'shuffle' option uses a Monte-Carlo approach,
        calculating ES or ECA scores for surrogate event time series obtained
        by shuffling the original event time series. The significance levels
        are the empirical percentiles of the ES/ECA scores of the original
        event series compared with the scores of the surrogate data.

        :type method: str 'ES' or 'ECA'
        :arg method: determines if ES or ECA should be used
        :type surrogate: str 'analytic' or 'shuffle'
        :arg surrogate: determines if p-values should be calculated using a
                        Monte-Carlo method or (only for ECA) an analytic
                        Poisson process null model
        :type n_surr: int
        :arg n_surr: number of surrogates for Monte-Carlo method
        :type symmetrization: str {'directed', 'symmetric', 'antisym',
                              'mean', 'max', 'min'} for ES,
                              str {'directed', 'mean', 'max', 'min'} for ECA
        :arg symmetrization: determines if and which symmetrisation
                             method should be used for the ES/ECA score matrix
        :type window_type: str {'retarded', 'advanced', 'symmetric'}
        :arg window_type: Only for ECA. Determines if precursor coincidence
                          rate ('advanced'), trigger coincidence rate
                          ('retarded') or a general coincidence rate with the
                          symmetric interval [-taumax, taumax] are computed
                          ('symmetric'). Default: 'symmetric'
        :rtype: 2D numpy array
        :return: significance levels of the ES/ECA scores for all pairs of
                 event series in event matrix
        """

        if method not in ['ES', 'ECA']:
            raise IOError("'method' parameter must be 'ECA' or 'ES'!")

        if surrogate not in ['analytic', 'shuffle']:
            raise IOError("'surrogate' parameter must be 'analytic' or "
                          "'shuffle'!")

        # Get instance variables
        deltaT = self.__taumax
        lag = self.__lag

        if method == 'ECA':

            if symmetrization not in ['directed', 'mean', 'max', 'min']:
                raise IOError("'symmetrization' parameter must be 'directed', "
                              "'mean', 'max' or 'min' for event"
                              "coincidence analysis!")

            if window_type not in ['retarded', 'advanced', 'symmetric']:
                raise IOError("'window_type' must be 'retarded',"
                              "'advanced' or 'symmetric'!")

            if surrogate == 'analytic':

                if symmetrization != 'directed':
                    raise IOError("'symmetrization' parameter should be"
                                  "'directed' for analytical calculation of"
                                  "significance levels!")

                if window_type not in ['retarded', 'advanced']:
                    raise IOError("'window_type' parameter must be 'retarded'"
                                  " or 'advanced' for analytical computation"
                                  " of significance levels!")

                # Compute ECA scores of stored event matrix
                directedECAMatrix = \
                    self._ndim_event_coincidence_analysis(
                        window_type=window_type)
                significance_levels = np.zeros((self.__N, self.__N))

                NEvents = self.__nrofevents

                if window_type == 'advanced':
                    for i in range(self.__N):
                        for j in range(i + 1, self.__N):
                            # Compute Poisson probability p
                            p = deltaT / (float(self.__timestamps[-1]) - lag)

                            # Compute number of precursor coincidences 2->1
                            K12 = int(directedECAMatrix[i][j] * NEvents[i])
                            # Compute probability of at least K12 precursor
                            # events
                            pvalue = 0.0
                            for K_star in range(K12, NEvents[i] + 1):
                                pvalue += \
                                    stats.binom.pmf(K_star, NEvents[i],
                                                    1.0 - pow(1.0 - p,
                                                              NEvents[j]))
                            significance_levels[i][j] = 1.0 - pvalue

                            # Compute number of precursor coincidence 1->2
                            K21 = int(directedECAMatrix[j][i] * NEvents[j])
                            # Compute probability of at least K21 precursor
                            # events
                            pvalue = 0.0
                            for K_star in range(K21, NEvents[j] + 1):
                                pvalue += \
                                    stats.binom.pmf(K_star, NEvents[j],
                                                    1.0 - pow(1.0 - p,
                                                              NEvents[i]))
                            significance_levels[j][i] = 1.0 - pvalue

                    return significance_levels

                # If window_type != 'advanced', it must be 'retarded'
                else:
                    for i in range(self.__N):
                        for j in range(i + 1, self.__N):
                            p = deltaT / (float(self.__timestamps[-1]) - lag)
                            # Compute probability of at least K12 trigger
                            # events

                            # Compute number of trigger coincidence 2->1
                            K12 = int(directedECAMatrix[i][j] * NEvents[j])
                            # Compute Poisson probability p
                            pvalue = 0.0
                            for K_star in range(K12, NEvents[j]):
                                pvalue += \
                                    stats.binom.pmf(K_star, NEvents[j],
                                                    1.0 - pow(1.0 - p,
                                                              NEvents[i]))
                            significance_levels[i][j] = 1.0 - pvalue

                            # Compute number of trigger coincidence 1->2
                            K21 = int(directedECAMatrix[j][i] * NEvents[i])
                            # Compute probability of at least K21 trigger
                            # events
                            pvalue = 0.0
                            for K_star in range(K21, NEvents[i]):
                                pvalue += \
                                    stats.binom.pmf(K_star, NEvents[i],
                                                    1.0 - pow(1.0 - p,
                                                              NEvents[j]))
                            significance_levels[j][i] = 1.0 - pvalue

                    return significance_levels

            # If surrogate is not 'analytic', it must be 'shuffle'
            else:
                return \
                    self._empirical_percentiles(method='ECA', n_surr=n_surr,
                                                symmetrization=symmetrization,
                                                window_type=window_type)

        elif method == 'ES':

            if surrogate != 'shuffle':
                raise IOError("Analytical calculation of significance level is"
                              " only possible for event coincidence analysis!")

            if symmetrization not in ['directed', 'symmetric', 'antisym',
                                      'mean', 'max', 'min']:
                raise IOError("'symmetrization' parameter must be 'directed', "
                              "'symmetric', 'antisym', 'mean', 'max' or"
                              "'min' for event synchronization!")

            return \
                self._empirical_percentiles(method='ES',
                                            n_surr=n_surr,
                                            symmetrization=symmetrization)
        else:
            return None
